import logging
import typing as t

import matplotlib.figure
import matplotlib.pyplot as plt
import mlflow
import numpy as np
import pandas as pd
import scipy.stats as st
import seaborn as sns
import sklearn.metrics

from . import evaluation

LOGGER = logging.getLogger(__name__)

# Z-score for 95% confidence interval
Z = st.norm.ppf(1 - (1 - 0.95) / 2)

# Flag FNR difference thresholds
HIGH_FLAG_THRESHOLD = 0.15
MODERATE_FLAG_THRESHOLD = 0.1
LOW_FLAG_THRESHOLD = 0.05

# Define flag types
FLAG_NAMES = {
    "⚪ INSUFFICIENT DATA": "insufficient_data",
    "🟢 NO BIAS": "no_bias",
    "🟡 LOW BIAS": "low_bias",
    "🟠 MODERATE BIAS": "moderate_bias",
    "🔴 HIGH BIAS": "high_bias",
}

# TODO: eventually we should use the custom_style.mplstyle colors, but currently
# the color palette does not have distinct enough colors for calibration by group
# where there are a lot of subgroups
PALETTE = sns.color_palette("Paired")

PosLabelType = t.Optional[int | float | bool | str]


def evaluate_bias(
    df_pred: pd.DataFrame,
    *,
    student_group_cols: list,
    pos_label: PosLabelType,
    split_col: str = "split",
    target_col: str = "target",
    pred_col: str = "pred",
    pred_prob_col: str = "pred_prob",
    sample_weight_col: str = "sample_weight_col",
) -> None:
    """
    Evaluates the bias in a model's predictions across different student groups for a split
    denoted by "split_name" across df_pred. For each student group, FNR (False Negative Rate)
    Parity is computed using absolute FNR percentage differences and any detected biases are
    flagged using hypothesis testing. Then, the metrics & plots are logged to MLflow.

    Args:
        run_id: The ID of the MLflow run
        df_pred: Pandas DataFrame with predictions from the model. The following columns
        need to be specified in df_pred: split_col, target_col, pred_col, and pred_probs_col.
        split_col: Column indicating split column ("train", "test", or "val")
        student_group_cols: A list of columns representing student groups for bias analysis
        target_col: Column name for the target (actual) values
        pred_col: Column name for the model's predicted values
        pred_prob_col: Column name for the model's predicted probabilities
        pos_label: Label representing the positive class
        sample_weight_col: Column name representing sample weights for the model.

    References:
        [1] https://fidelity.github.io/jurity/about_fairness.html
        [2] https://docs.oracle.com/en-us/iaas/tools/automlx/latest/legacy/v23.2.3/fairness.html
        [3] https://search.r-project.org/CRAN/refmans/fairness/html/fnr_parity.html
    """
    model_flags = []

    for split_name, split_data in df_pred.groupby(split_col):
        for group_col in student_group_cols:
            bias_metrics, perf_metrics, fnr_data = compute_group_bias_metrics(
                split_data,
                split_name,
                group_col,
                target_col,
                pred_col,
                pred_prob_col,
                pos_label,
                sample_weight_col,
            )

            log_group_metrics_to_mlflow(bias_metrics, f"bias_{split_name}", group_col)
            log_group_metrics_to_mlflow(perf_metrics, f"perf_{split_name}", group_col)

            # Detect bias flags
            all_flags = flag_bias(fnr_data)

            # Filter flags for groups where bias is detected
            group_flags = [
                flag
                for flag in all_flags
                if flag["flag"] not in ["🟢 NO BIAS", "⚪ INSUFFICIENT DATA"]
            ]

            if group_flags:
                fnr_fig = plot_fnr_group(fnr_data)
                mlflow.log_figure(
                    fnr_fig, f"fnr_plots/{split_name}_{group_col}_fnr.png"
                )
                plt.close()

                for flag in group_flags:
                    LOGGER.warning(
                        " Bias detected for %s on %s - %s, FNR Difference: %.2f%% (%s) [%s]",
                        flag["group"],
                        flag["split_name"],
                        flag["subgroups"],
                        flag["fnr_percentage_difference"] * 100,
                        flag["type"],
                        flag["flag"],
                    )

            model_flags.extend(all_flags)

    log_bias_flags_to_mlflow(model_flags)


def compute_group_bias_metrics(
    split_data: pd.DataFrame,
    split_name: str,
    group_col: str,
    target_col: str,
    pred_col: str,
    pred_prob_col: str,
    pos_label: PosLabelType,
    sample_weight_col: str,
) -> tuple[list, list, list]:
    """
    Computes group metrics (including FNR) based on evaluation parameters and logs them to MLflow.
    We split bias & performance metrics into separate dictionaries in our output to make the model
    card more readable with two separate tables.
    Args:
        split_data: Data for the current split to evaluate
        split_name: Name of the data split (e.g., "train", "test", or "val")
        target_col: Column name for the target variable.
        pred_col: Column name for the predictions.
        pred_prob_col: Column name for predicted probabilities.
        pos_label: Positive class label.
        student_group_cols: List of columns for subgroups.
    """
    bias_group_metrics = []
    perf_group_metrics = []
    fnr_data = []

    for subgroup_name, subgroup_data in split_data.groupby(group_col):
        labels = subgroup_data[target_col]
        preds = subgroup_data[pred_col]
        pred_probs = subgroup_data[pred_prob_col]

        fnr, fnr_lower, fnr_upper, num_positives = calculate_fnr_and_ci(labels, preds)

        fnr_subgroup_data = {
            "group": group_col,
            "subgroup": subgroup_name,
            "fnr": fnr,
            "split_name": split_name,
            "ci": (fnr_lower, fnr_upper),
            "size": len(subgroup_data),
            "number_of_positive_samples": num_positives,
        }

        eval_metrics = evaluation.compute_classification_perf_metrics(
            labels,
            preds,
            pred_probs,
            pos_label=pos_label,
            sample_weights=(
                subgroup_data[sample_weight_col]
                if sample_weight_col in subgroup_data.columns
                else None
            ),
        )
        # HACK: avoid duplicative metrics
        eval_metrics.pop("num_positives", None)
        bias_subgroup_metrics, perf_subgroup_metrics = format_subgroup_metrics(
            eval_metrics, fnr_subgroup_data
        )

        log_subgroup_metrics_to_mlflow(
            bias_subgroup_metrics, f"{split_name}_bias", group_col
        )
        log_subgroup_metrics_to_mlflow(
            perf_subgroup_metrics, f"{split_name}_performance", group_col
        )

        bias_group_metrics.append(bias_subgroup_metrics)
        perf_group_metrics.append(perf_subgroup_metrics)
        fnr_data.append(fnr_subgroup_data)

    return bias_group_metrics, perf_group_metrics, fnr_data


def format_subgroup_metrics(
    eval_metrics: dict, fnr_subgroup_data: dict
) -> tuple[dict, dict]:
    """
    Formats the evaluation metrics and bias metrics together for logging into MLflow.

    Args:
        eval_metrics: Dictionary performance metrics for subgroup
        fnr_subgroup_data: List of dictionaries containing FNR and CI information for each subgroup.
    Returns:
        Tuple of dictionaries summarizing both performance & bias metrics
    """
    bias_metrics = {
        "Subgroup": fnr_subgroup_data["subgroup"],
        "Number of Samples": eval_metrics["num_samples"],
        "Number of Positive Samples": fnr_subgroup_data["number_of_positive_samples"],
        "Actual Target Prevalence": round(eval_metrics["true_positive_prevalence"], 2),
        "Predicted Target Prevalence": round(
            eval_metrics["pred_positive_prevalence"], 2
        ),
        # Bias Metrics
        "FNR": round(fnr_subgroup_data["fnr"], 2),
        "FNR CI Lower": round(fnr_subgroup_data["ci"][0], 2),
        "FNR CI Upper": round(fnr_subgroup_data["ci"][1], 2),
    }
    performance_metrics = {
        "Subgroup": fnr_subgroup_data["subgroup"],
        # Performance Metrics
        "Accuracy": round(eval_metrics["accuracy"], 2),
        "Precision": round(eval_metrics["precision"], 2),
        "Recall": round(eval_metrics["recall"], 2),
        "F1 Score": round(eval_metrics["f1_score"], 2),
        "Log Loss": round(eval_metrics["log_loss"], 2),
    }
    return bias_metrics, performance_metrics


def flag_bias(
    fnr_data: list,
    high_bias_thresh: float = HIGH_FLAG_THRESHOLD,
    moderate_bias_thresh: float = MODERATE_FLAG_THRESHOLD,
    low_bias_thresh: float = LOW_FLAG_THRESHOLD,
    min_sample_ratio: float = 0.15,
) -> list[dict]:
    """
    Flags bias based on FNR Parity (via absolute FNR percentage differences) and confidence interval overlap.

    Args:
        fnr_data: List of dictionaries containing FNR and CI information for each subgroup.
        high_bias_thresh: Threshold for flagging high bias.
        moderate_bias_thresh: Threshold for flagging moderate bias.
        low_bias_thresh: Threshold for flagging low bias.
        min_sample_ratio: Percentage of total positive samples required for valid FNR comparison.
        This gives us flexibility with smaller datasets. We default to 15% since we want to ensure
        we are checking subgroups with sufficient data. When calculating min_samples, we have an upper
        limit of 50 samples so that larger datasets aren't unnecessarily restricted.

    Returns:
        List of dictionaries with bias flag information.
    """
    total_group_positives = sum(
        subgroup["number_of_positive_samples"] for subgroup in fnr_data
    )
    min_samples = min(50, int(min_sample_ratio * total_group_positives))

    bias_flags = []
    thresholds = [
        (high_bias_thresh, "🔴 HIGH BIAS", 0.01),
        (moderate_bias_thresh, "🟠 MODERATE BIAS", 0.01),
        (low_bias_thresh, "🟡 LOW BIAS", 0.1),
    ]

    for i, current in enumerate(fnr_data):
        for other in fnr_data[i + 1 :]:
            if current["fnr"] > 0 and other["fnr"] > 0:
                # Determine ordering based on FNR values
                sg1, sg2 = (
                    (current, other)
                    if current["fnr"] >= other["fnr"]
                    else (other, current)
                )
                # Guaranteed to be greater than zero
                fnr_diff = sg1["fnr"] - sg2["fnr"]
                p_value = z_test_fnr_difference(
                    sg1["fnr"], sg2["fnr"], sg1["size"], sg2["size"]
                )
                ci_overlap = check_ci_overlap(sg1["ci"], sg2["ci"])

                if np.isnan(p_value) or (
                    (sg1["number_of_positive_samples"] < min_samples)
                    or (sg2["number_of_positive_samples"] < min_samples)
                ):
                    bias_flags.append(
                        generate_bias_flag(
                            sg1["group"],
                            sg1["subgroup"],
                            sg2["subgroup"],
                            fnr_diff,
                            "insufficient samples for statistical test",
                            sg1["split_name"],
                            "⚪ INSUFFICIENT DATA",
                            p_value,
                        )
                    )
                elif fnr_diff < low_bias_thresh or p_value > 0.1:
                    bias_flags.append(
                        generate_bias_flag(
                            sg1["group"],
                            sg1["subgroup"],
                            sg2["subgroup"],
                            fnr_diff,
                            "no significant difference",
                            sg1["split_name"],
                            "🟢 NO BIAS",
                            p_value,
                        )
                    )
                else:
                    for threshold, flag, p_thresh in thresholds:
                        if fnr_diff >= threshold and p_value <= p_thresh:
                            reason = (
                                "overlapping confidence intervals"
                                if ci_overlap
                                else "non-overlapping confidence intervals"
                            )
                            bias_flags.append(
                                generate_bias_flag(
                                    sg1["group"],
                                    sg1["subgroup"],
                                    sg2["subgroup"],
                                    fnr_diff,
                                    reason,
                                    sg1["split_name"],
                                    flag,
                                    p_value,
                                )
                            )
                            break  # Exit after the first matched threshold

    return bias_flags


def calculate_fnr_and_ci(
    targets: pd.Series,
    preds: pd.Series,
    apply_scaling: bool = True,
) -> tuple[float, float, float, bool]:
    """
    Calculates the False Negative Rate (FNR) and its confidence interval, applying Log scaling.

    Args:
        targets: "Actual" labels from model output
        preds: Predictions from model output
        min_fnr_samples: Minimum number of true positives or false negatives for FNR calculation.
        apply_scaling: Boolean of whether log scaling should be applied. We default to True since we want to
        sufficiently dampen FNR variance in low sample sizes situations.

    Returns:
        fnr: False Negative Rate
        ci_min: Lower bound of the confidence interval
        ci_max: Upper bound of the confidence interval
        num_positives: Number of positive samples, for reporting. When the number of positives is low, the FNR computation may not be reliable.
    """
    cm = sklearn.metrics.confusion_matrix(targets, preds, labels=[False, True])
    tn, fp, fn, tp = cm.ravel()

    # Calculate FNR & apply Log Scaling to smoothen FNR at low sample sizes
    num_positives = fn + tp
    if apply_scaling:
        num_positives += np.log1p(num_positives)

    fnr = fn / num_positives if num_positives > 0 else 0

    # Confidence Interval Calculation
    margin = Z * np.sqrt((fnr * (1 - fnr)) / num_positives) if num_positives > 0 else 0
    ci_min, ci_max = max(0, fnr - margin), min(1, fnr + margin)

    return fnr, ci_min, ci_max, fn + tp


def check_ci_overlap(
    ci1: tuple[float, float],
    ci2: tuple[float, float],
) -> bool:
    """
    Checks whether confidence intervals (CIs) overlap. If they do, the FNR differences
    are within the margin of error at the 95% confidence level. If the CIs do not
    overlap, this suggests strong statistical evidence that the FNRs are different.

    Args:
        ci1: Confidence interval (min, max) for subgroup 1
        ci2: Confidence interval (min, max) for subgroup 2

    Returns:
        Boolean indicating whether the CIs overlap.
    """
    return not (ci1[1] < ci2[0] or ci2[1] < ci1[0])


def z_test_fnr_difference(
    fnr1: float,
    fnr2: float,
    num_positives1: int,
    num_positives2: int,
) -> float:
    """
    Performs a z-test for the FNR difference between two groups. If there are
    less than 30 samples of false negatives and true negatives, then we do not
    have enough data to perform a z-test. Thirty samples is the standard check
    for z-tests.

    Args:
        fnr1: FNR value for subgroup 1
        fnr2: FNR value for subgroup 2
        num_positives1: Number of false negatives + true negatives for subgroup 1
        num_positives2: Number of false negatives + true negatives for subgroup 2

    Returns:
        Two-tailed p-value for the z-test for the raw FNR difference between the two subgroups.
    """
    if (
        num_positives1 <= 30 or num_positives2 <= 30
    ):  # Ensures valid sample sizes for z-test
        return np.nan
    std_error = np.sqrt(
        ((fnr1 * (1 - fnr1)) / num_positives1) + ((fnr2 * (1 - fnr2)) / num_positives2)
    )
    z_stat = (fnr1 - fnr2) / std_error
    return float(2 * (1 - st.norm.cdf(abs(z_stat))))  # Two-tailed p-value


def generate_bias_flag(
    group: str,
    subgroup1: str,
    subgroup2: str,
    fnr_percentage_difference: float,
    bias_type: str,
    split_name: str,
    flag: str,
    p_value: float = np.nan,
) -> dict:
    """
    Aggregate bias flag information for a given subgroup pair into a dict.

    Args:
        group: Name of the group (e.g. "Gender", "Race", "Age")
        subgroup1: Name of the subgroup 1 (e.g. "Female", "Male", "Asian", "African American", "Caucasian")
        subgroup2: Name of the subgroup 2 (e.g. "Female", "Male", "Asian", "African American", "Caucasian")
        fnr_percentage_difference: Absolute value of percentage difference in FNR
        bias_type: Type of bias (e.g. "Non-overlapping CIs", "Overlapping: : p-value: ...")
        split_name: Name of the split (e.g. train/test/validate)
        flag: Flag value (e.g. "🔴 HIGH BIAS", "🟠 MODERATE BIAS", "🟡 LOW BIAS", "🟢 NO BIAS")
        p_value: p-value for the z-test for the FNR difference of the subgroup pair

    Returns:
        Dictionary containing bias flag information.
    """
    flag_entry = {
        "group": group,
        "subgroups": f"{subgroup1} vs {subgroup2}",
        "fnr_percentage_difference": round(fnr_percentage_difference, 4),
        "type": (
            bias_type
            if np.isnan(p_value)
            else f"{bias_type} with a p-value {'less than 0.001' if p_value < 0.001 else f'of {p_value:.3f}'}"
        ),
        "split_name": split_name,
        "flag": flag,
    }
    return flag_entry


def log_bias_flags_to_mlflow(all_model_flags: list) -> None:
    """
    Save and log bias flags to MLflow. If no flags exist for the model, then we do not log anything.

    Args:
        all_model_flags: Bias flags for across all splits
        (e.g. "train", "test", "val") of the model
    """
    if all_model_flags:
        df_model_flags = pd.DataFrame(all_model_flags)
        for flag in FLAG_NAMES.keys():
            flag_name = FLAG_NAMES[flag]
            df_flag = (
                df_model_flags[df_model_flags["flag"] == flag].sort_values(
                    by="fnr_percentage_difference", ascending=False
                )
                if df_model_flags.shape[0] > 0
                else None
            )
            if df_flag is not None:
                bias_tmp_path = f"/tmp/{flag_name}_flags.csv"
                df_flag.to_csv(bias_tmp_path, index=False)
                mlflow.log_artifact(
                    local_path=bias_tmp_path, artifact_path="bias_flags"
                )


def log_group_metrics_to_mlflow(
    group_metrics: list,
    split_name: str,
    group_col: str,
) -> None:
    """
    Saves and logs group-level bias metrics as a CSV artifact in MLflow.

    Args:
        group_metrics: List of dictionaries containing computed group-level bias
        or performance metrics.
        split_name: Name of the data split (e.g., "train", "test", "validation").
        group_col: Column name representing the group for bias evaluation.
    """
    df_group_metrics = pd.DataFrame(group_metrics)
    metrics_tmp_path = f"/tmp/{split_name}_{group_col}_metrics.csv"
    df_group_metrics.to_csv(metrics_tmp_path, index=False)
    mlflow.log_artifact(local_path=metrics_tmp_path, artifact_path="group_metrics")


def log_subgroup_metrics_to_mlflow(
    subgroup_metrics: dict,
    split_name: str,
    group_col: str,
) -> None:
    """
    Logs individual subgroup-level metrics to MLflow.

    Args:
        subgroup_metrics: Dictionary of subgroup bias metrics.
        split_name: Name of the data split (e.g., "train", "test", "validation").
        group_col: Column name representing the group for bias evaluation.
    """
    for metric, value in subgroup_metrics.items():
        if metric not in {"Subgroup", "Number of Samples"}:
            mlflow.log_metric(
                f"{split_name}_{group_col}_metrics/{metric}_subgroup", value
            )


def plot_fnr_group(fnr_data: list) -> matplotlib.figure.Figure:
    """
    Plots False Negative Rate (FNR) for a group by subgroup on
    a split (train/test/val) of data with confidence intervals.

    Args:
        fnr_data: List with dictionaries for each subgroup. Each dictionary
        comprises of group, fnr, ci (confidence interval), and split_name keys.
    """
    subgroups = [subgroup_data["subgroup"] for subgroup_data in fnr_data]
    fnr = [subgroup_data["fnr"] for subgroup_data in fnr_data]
    min_ci_errors = [
        subgroup_data["fnr"] - subgroup_data["ci"][0] for subgroup_data in fnr_data
    ]
    max_ci_errors = [
        subgroup_data["ci"][1] - subgroup_data["fnr"] for subgroup_data in fnr_data
    ]

    y_positions = list(range(len(subgroups)))
    fig, ax = plt.subplots()

    for i, (x, y, err_min, err_max) in enumerate(
        zip(fnr, y_positions, min_ci_errors, max_ci_errors)
    ):
        color = PALETTE[i % len(PALETTE)]
        ax.errorbar(
            x=x,
            y=y,
            xerr=[[err_min], [err_max]],
            fmt="o",
            capsize=5,
            capthick=2,
            markersize=8,
            linewidth=2,
            color=color,
            label="FNR" if i == 0 else "",
        )

    ax.set(
        yticks=y_positions,
        yticklabels=subgroups,
        ylim=(-1, len(subgroups)),
        ylabel="Subgroup",
        xlabel="False Negative Rate",
        title=f"""FNR @ 0.5 for {fnr_data[0]["group"]} on {fnr_data[0]["split_name"]}""",
    )
    ax.tick_params(axis="both", labelsize=12)
    ax.grid(True, linestyle="--", alpha=0.6)

    plt.tight_layout()
    return fig



import re
import typing as t

import pydantic as pyd


class CustomProjectConfig(pyd.BaseModel):
    """Configuration schema for SST Custom projects."""

    institution_id: str = pyd.Field(
        ...,
        description=(
            "Unique (ASCII-only) identifier for institution; used in naming things "
            "such as source directories, catalog schemas, keys in shared configs, etc."
        ),
    )
    institution_name: str = pyd.Field(
        ...,
        description=(
            "Readable 'display' name for institution, distinct from the 'id'; "
            "probably just the school's 'official', public name"
        ),
    )

    # shared parameters
    student_id_col: str = "student_id"
    target_col: str = "target"
    split_col: str = "split"
    sample_weight_col: t.Optional[str] = "sample_weight"
    student_group_cols: t.Optional[list[str]] = pyd.Field(
        default=["age_group", "race", "gender"],
        description=(
            "One or more column names in datasets containing student 'groups' "
            "to use for model bias assessment, but *not* as model features"
        ),
    )
    student_group_aliases: dict[str, str] = pyd.Field(
        default_factory=dict,
        description=(
            "Mapping from raw column name (e.g., GENDER_DESC) to "
            "friendly label (e.g., 'Gender') for use in model cards"
        ),
    )
    pred_col: str = "pred"
    pred_prob_col: str = "pred_prob"
    pos_label: t.Optional[int | bool | str] = True
    random_state: t.Optional[int] = 12345

    # key artifacts produced by project pipeline
    datasets: "AllDatasetStagesConfig" = pyd.Field(
        description=(
            "Key datasets produced by the pipeline represented here in this config"
        ),
    )
    model: t.Optional["ModelConfig"] = pyd.Field(
        default=None,
        description=(
            "Essential metadata for identifying and loading trained model artifacts, "
            "as produced by the pipeline represented here in this config"
        ),
    )

    # key steps in project pipeline
    preprocessing: t.Optional["PreprocessingConfig"] = None
    modeling: t.Optional["ModelingConfig"] = None
    inference: t.Optional["InferenceConfig"] = None

    @pyd.computed_field  # type: ignore[misc]
    @property
    def non_feature_cols(self) -> list[str]:
        return (
            [self.student_id_col, self.target_col]
            + ([self.split_col] if self.split_col else [])
            + ([self.sample_weight_col] if self.sample_weight_col else [])
            + (self.student_group_cols or [])
        )

    @pyd.field_validator("institution_id", mode="after")
    @classmethod
    def check_institution_id_isascii(cls, value: str) -> str:
        if not re.search(r"^\w+$", value, flags=re.ASCII):
            raise ValueError(f"institution_id='{value}' is not ASCII-only")
        return value

    # NOTE: this is for *pydantic* model -- not ML model -- configuration
    model_config = pyd.ConfigDict(extra="forbid", strict=True)

    @pyd.model_validator(mode="after")
    def check_sample_weight_requires_random_state(self):
        if self.sample_weight_col and self.random_state is None:
            raise ValueError(
                "random_state must be specified if sample_weight_col is provided"
            )
        return self

    @pyd.model_validator(mode="after")
    def validate_student_group_aliases(self) -> "CustomProjectConfig":
        missing = [
            col
            for col in (self.student_group_cols or [])
            if col not in (self.student_group_aliases or {})
        ]
        if missing:
            raise ValueError(f"Missing student_group_aliases for: {missing}")
        return self


class DatasetConfig(pyd.BaseModel):
    train_file_path: t.Optional[str] = pyd.Field(
        default=None,
        description="Absolute path to training dataset on disk.",
    )
    predict_file_path: t.Optional[str] = pyd.Field(
        default=None,
        description="Absolute path to prediction/inference dataset on disk.",
    )
    train_table_path: t.Optional[str] = pyd.Field(
        default=None,
        description="Unity Catalog table path for training dataset, e.g., 'catalog.schema.table'.",
    )
    predict_table_path: t.Optional[str] = pyd.Field(
        default=None,
        description="Unity Catalog table path for prediction/inference dataset.",
    )
    file_path: t.Optional[str] = None
    table_path: t.Optional[str] = None

    primary_keys: t.Optional[t.List[str]] = pyd.Field(
        default=None,
        description="Primary keys utilized for data validation, if applicable",
    )
    drop_cols: t.Optional[t.List[str]] = pyd.Field(
        default=None,
        description="Columns to be dropped during pre-processing, if applicable",
    )
    non_null_cols: t.Optional[t.List[str]] = pyd.Field(
        default=None,
        description="Columns to be validated as non-null, if applicable",
    )

    @pyd.model_validator(mode="after")
    def validate_paths(self) -> "DatasetConfig":
        any_paths = [
            self.train_file_path,
            self.predict_file_path,
            self.train_table_path,
            self.predict_table_path,
            self.file_path,  # Legacy, not used in pipeline/DB workflow
            self.table_path,  # Legacy, not used in pipeline/DB workflow
        ]
        if not any(any_paths):
            raise ValueError(
                "At least one dataset path must be specified: "
                "`train_file_path`, `predict_file_path`, "
                "`train_table_path`, `predict_table_path`, "
                "`file_path`, or `table_path`"
            )
        return self

    def get_path(self, mode: t.Literal["train", "predict"]) -> t.Optional[str]:
        """Convenience accessor for the train/predict path."""
        if mode == "train":
            return self.train_file_path or self.train_table_path
        elif mode == "predict":
            return self.predict_file_path or self.predict_table_path
        else:
            raise ValueError(f"Unknown mode: {mode}")


class AllDatasetStagesConfig(pyd.BaseModel):
    bronze: dict[str, DatasetConfig]
    silver: dict[str, DatasetConfig]
    gold: dict[str, DatasetConfig]


class ModelConfig(pyd.BaseModel):
    experiment_id: str
    run_id: str
    framework: t.Optional[t.Literal["sklearn", "xgboost", "lightgbm"]] = None

    @pyd.computed_field  # type: ignore[misc]
    @property
    def mlflow_model_uri(self) -> str:
        return f"runs:/{self.run_id}/model"


class PreprocessingConfig(pyd.BaseModel):
    selection: "SelectionConfig"
    checkpoint: "CheckpointConfig"
    target: "TargetConfig"
    splits: dict[t.Literal["train", "test", "validate"], float] = pyd.Field(
        default={"train": 0.6, "test": 0.2, "validate": 0.2},
        description=(
            "Mapping of name to fraction of the full datset belonging to a given 'split', "
            "which is a randomized subset used for different parts of the modeling process"
        ),
    )
    sample_class_weight: t.Optional[t.Literal["balanced"] | dict[object, int]] = (
        pyd.Field(
            default=None,
            description=(
                "Weights associated with classes in the form ``{class_label: weight}`` "
                "or 'balanced' to automatically adjust weights inversely proportional "
                "to class frequencies in the input data. "
                "If null (default), then sample weights are not computed."
            ),
        )
    )

    @pyd.field_validator("splits", mode="after")
    @classmethod
    def check_split_fractions(cls, value: dict) -> dict:
        if (sum_fracs := sum(value.values())) != 1.0:
            raise pyd.ValidationError(
                f"split fractions must sum up to 1.0, but input sums up to {sum_fracs}"
            )
        return value


class CheckpointConfig(pyd.BaseModel):
    name: str = pyd.Field(default="checkpoint")
    params: dict[str, object] = pyd.Field(default_factory=dict)
    unit: t.Literal["credit", "year", "term", "semester"]
    value: int = pyd.Field(
        default=30,
        description=(
            "Number of checkpoint units (e.g. 1 year, 1 term/semester, 30 credits)"
        ),
    )
    optional_desc: t.Optional[str] = pyd.Field(
        default=None,
        description=(
            "Optional description of the checkpoint beyond the unit and value. "
            "Used to provide further context for the particular institution and model. "
        ),
    )

    @pyd.field_validator("value")
    @classmethod
    def check_value_positive(cls, v):
        if v is not None and v <= 0:
            raise ValueError("Value must be greater than zero.")
        return v


class TargetConfig(pyd.BaseModel):
    name: str = pyd.Field(default="target")
    category: t.Literal["graduation", "retention"]
    params: dict[str, object] = pyd.Field(default_factory=dict)
    unit: t.Literal["credit", "year", "term", "semester", "pct_completion"]
    value: int = pyd.Field(
        default=120,
        description=(
            "Number of target units (e.g. 4 years, 4 terms, 120 credits, 150 completion %)"
        ),
    )
    optional_desc: t.Optional[str] = pyd.Field(
        default=None,
        description=(
            "Optional description of the target beyond the unit and value. "
            "Used to provide further context for the particular institution and model. "
        ),
    )

    @pyd.field_validator("value")
    @classmethod
    def check_value_positive(cls, v):
        if v is not None and v <= 0:
            raise ValueError("Value must be greater than zero.")
        return v


class SelectionConfig(pyd.BaseModel):
    student_criteria: dict[str, object] = pyd.Field(
        default_factory=dict,
        description=(
            "Column name in modeling dataset mapped to one or more values that it must equal "
            "in order for the corresponding student to be considered 'eligible'. "
            "Multiple criteria are combined with a logical 'AND'."
        ),
    )
    student_criteria_aliases: dict[str, str] = pyd.Field(
        default_factory=dict,
        description="Human-readable display names for student_criteria keys",
    )

    @pyd.model_validator(mode="after")
    def validate_criteria_aliases(self) -> "SelectionConfig":
        criteria_keys = self.student_criteria.keys()
        alias_keys = self.student_criteria_aliases or {}

        missing = [k for k in criteria_keys if k not in alias_keys]
        if missing:
            raise ValueError(
                f"Missing display aliases in `student_criteria_aliases` for: {missing}"
            )
        return self


class ModelingConfig(pyd.BaseModel):
    feature_selection: t.Optional["FeatureSelectionConfig"] = None
    training: "TrainingConfig"
    evaluation: t.Optional["EvaluationConfig"] = None
    bias_mitigation: t.Optional["BiasMitigationConfig"] = None


class FeatureSelectionConfig(pyd.BaseModel):
    """
    See Also:
        - :func:`modeling.feature_selection.select_features()`
    """

    force_include_cols: t.Optional[list[str]] = None
    incomplete_threshold: float = 0.5
    low_variance_threshold: float = 0.0
    collinear_threshold: t.Optional[float] = 10.0


class TrainingConfig(pyd.BaseModel):
    """
    References:
        - https://docs.databricks.com/en/machine-learning/automl/automl-api-reference.html#classify
    """

    exclude_cols: t.Optional[list[str]] = pyd.Field(
        default=None,
        description="One or more column names in dataset to exclude from training.",
    )
    time_col: t.Optional[str] = pyd.Field(
        default=None,
        description=(
            "Column name in dataset used to split train/test/validate sets chronologically, "
            "as an alternative to the randomized assignment in ``split_col`` ."
        ),
    )
    exclude_frameworks: t.Optional[list[str]] = pyd.Field(
        default=None,
        description="List of algorithm frameworks that AutoML excludes from training.",
    )
    primary_metric: str = pyd.Field(
        default="log_loss",
        description="Metric used to evaluate and rank model performance.",
    )
    timeout_minutes: t.Optional[int] = pyd.Field(
        default=None,
        description="Maximum time to wait for AutoML trials to complete.",
    )


class EvaluationConfig(pyd.BaseModel):
    topn_runs_included: int = pyd.Field(
        default=3,
        description="Number of top-scoring mlflow runs to include for detailed evaluation",
    )


class BiasMitigationConfig(pyd.BaseModel):
    student_group_col: str = pyd.Field(
        default="student_group",
        description="Column name in dataset to have a custom threshold set.",
    )
    student_group_col_alias: str = pyd.Field(
        default="Student Group",
        description="Human-readable display name for student_group column.",
    )
    student_group: str = pyd.Field(
        default="freshmen",
        description="Value in student_group column that has a custom threshold based on bias considerations.",
    )
    custom_threshold: float = pyd.Field(
        default=0.5,
        description="Threshold for student group based on bias considerations.",
    )


class InferenceConfig(pyd.BaseModel):
    num_top_features: int = pyd.Field(default=5)
    min_prob_pos_label: t.Optional[float] = 0.5



import logging
import pathlib
import typing as t

import pandas as pd
import pyspark.sql

try:
    import tomllib  # noqa
except ImportError:  # => PY3.10
    import tomli as tomllib  # noqa

LOGGER = logging.getLogger(__name__)


def from_csv_file(
    file_path: str,
    spark_session: t.Optional[pyspark.sql.SparkSession] = None,
    **kwargs: object,
) -> pd.DataFrame:
    """
    Read data from a CSV file at ``file_path`` and return it as a DataFrame.

    Args:
        file_path: Path to file on disk from which data will be read.
        spark_session: If given, data is loaded via ``pyspark.sql.DataFrameReader.csv`` ;
            otherwise, data is loaded via :func:`pandas.read_csv()` .
        **kwargs: Additional arguments passed as-is into underlying read func.

    See Also:
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.csv.html
    """
    if spark_session is None:
        df = pd.read_csv(file_path, dtype="string", header="infer", **kwargs)  # type: ignore
    else:
        df = spark_session.read.csv(
            file_path,
            inferSchema=False,
            header=True,
            **kwargs,  # type: ignore
        ).toPandas()
    assert isinstance(df, pd.DataFrame)  # type guard
    LOGGER.info("loaded rows x cols = %s from '%s'", df.shape, file_path)
    return df


def from_delta_table(
    table_path: str, spark_session: pyspark.sql.SparkSession
) -> pd.DataFrame:
    """
    Read data from a table in Databricks Unity Catalog and return it as a DataFrame.

    Args:
        table_path: Path in Unity Catalog from which data will be read,
            including the full three-level namespace: ``catalog.schema.table`` .
        spark_session: Entry point to using spark dataframes and the databricks integration.

    See Also:
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.html#pyspark.sql.DataFrameReader
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/spark_session.html
    """
    df = spark_session.read.format("delta").table(table_path).toPandas()
    assert isinstance(df, pd.DataFrame)  # type guard
    LOGGER.info("loaded rows x cols = %s of data from '%s'", df.shape, table_path)
    return df


def from_toml_file(file_path: str) -> dict[str, object]:
    """
    Read data from ``file_path`` and return it as a dict.

    Args:
        file_path: Path to file on disk from which data will be read.
    """
    fpath = pathlib.Path(file_path).resolve()
    with fpath.open(mode="rb") as f:
        data = tomllib.load(f)
    LOGGER.info("loaded config from '%s'", fpath)
    assert isinstance(data, dict)  # type guard
    return data



import logging
import typing as t
import pandas as pd
import collections

from ..utils import utils

LOGGER = logging.getLogger(__name__)


def register_evaluation_sections(card, registry):
    """
    Register evaluation metrics for each group. These metrics include both performance and bias.
    We assume all necessary formatting in terms of rows and columns of the table is done in the
    mlflow artifact.
    """
    performance_section = [f"{card.format.header_level(4)}Model Performance\n"]
    split_artifacts = utils.list_paths_in_directory(
        run_id=card.run_id, directory="metrics"
    )

    evaluation_sections = [
        f"{card.format.header_level(4)}Evaluation Metrics by Student Group\n"
    ]
    group_eval_artifacts = utils.list_paths_in_directory(
        run_id=card.run_id, directory="group_metrics"
    )

    def make_metric_table(path: str, title: str) -> t.Callable[[], str]:
        """
        This method is used for dynamic section registration.
        Later, the registry will render all of these functions to create
        tables.

        Args:
            path: Artifact path to the csv file containing the evaluation metrics.
            title: Title of the group.

        Returns:
            A function that returns a markdown table of the evaluation metrics for a group.
        """

        def metric_table():
            try:
                local_path = utils.download_artifact(
                    run_id=card.run_id,
                    local_folder=card.assets_folder,
                    artifact_path=path,
                )
                df = pd.read_csv(local_path)

                # Build markdown table
                headers = "| " + " | ".join(df.columns) + " |"
                separator = "| " + " | ".join(["---"] * len(df.columns)) + " |"
                rows = [
                    "| " + " | ".join(str(val) for val in row) + " |"
                    for row in df.values
                ]

                return f"{card.format.header_level(5)}{title}\n\n" + "\n".join(
                    [headers, separator] + rows
                )

            except Exception as e:
                LOGGER.warning(
                    f"Could not load evaluation metrics for {title}: {str(e)}"
                )
                return f"{card.format.bold(f'{title} Metrics')}\n\nCould not load data."

        return metric_table

    # Evaluation by Group
    group_parts = collections.defaultdict(dict)

    # Group bias and performance parts for each group
    for csv_path in group_eval_artifacts:
        if csv_path.startswith("group_metrics/bias_test_") and csv_path.endswith(
            ".csv"
        ):
            group_name = csv_path.replace("group_metrics/bias_test_", "").replace(
                "_metrics.csv", ""
            )
            group_parts[group_name]["bias"] = csv_path

        if csv_path.startswith("group_metrics/perf_test_") and csv_path.endswith(
            ".csv"
        ):
            group_name = csv_path.replace("group_metrics/perf_test_", "").replace(
                "_metrics.csv", ""
            )
            group_parts[group_name]["perf"] = csv_path

    # Render both tables under the same group title without labeling them separately
    for group_name, parts in group_parts.items():
        if "bias" not in parts or "perf" not in parts:
            continue

        section_text = []

        # Bias Evaluation Table
        bias_title = f"{card.format.friendly_case(group_name)} Bias Metrics"
        bias_table_func = make_metric_table(parts["bias"], bias_title)
        registry.register(f"metric_table_{group_name}_bias")(bias_table_func)
        section_text.append(bias_table_func())

        # Performance Evaluation Table
        perf_title = f"{card.format.friendly_case(group_name)} Performance Metrics"
        perf_table_func = make_metric_table(parts["perf"], perf_title)
        registry.register(f"metric_table_{group_name}_perf")(perf_table_func)
        section_text.append(perf_table_func())

        evaluation_sections.append("\n\n".join(section_text))

    # Performance Across Splits
    for csv_path in split_artifacts:
        if csv_path.startswith("metrics/") and csv_path.endswith("_splits.csv"):
            title = "Performance across Splits"
            table_func = make_metric_table(csv_path, title)
            registry.register("performance_by_splits_metric_table")(table_func)
            performance_section.append(table_func())

    @registry.register("evaluation_by_group_section")
    def evaluation_section():
        """
        Returns the evaluation metrics section for the model card.
        """
        if not evaluation_sections:
            return f"{card.format.bold('No group evaluation metrics available')}."
        return "\n\n".join(evaluation_sections)

    @registry.register("performance_by_splits_section")
    def performance_by_splits_section():
        """
        Returns the performance metrics section by splits for the model card.
        """
        if not performance_section:
            return f"{card.format.bold('No performance metrics available')}."
        return "\n\n".join(performance_section)



import logging
import pandas as pd

from ..utils import utils

LOGGER = logging.getLogger(__name__)


def register_data_sections(card, registry):
    """
    Registers sections related to data split and feature tables.
    It does not contain sections relevant to performance or bias. These tables
    are gathered from a combination of config.toml and mlflow artifacts.
    """

    @registry.register("data_split_table")
    def data_split():
        """
        Produces a markdown table of the data split.
        """
        total_students = len(card.modeling_data)
        splits = card.cfg.preprocessing.splits
        if splits:
            labels = {"train": "Training", "validate": "Validation", "test": "Test"}

            rows = [
                f"| {labels[k]:<10} | {round(total_students * splits[k]):<8} | {int(splits[k] * 100)}%       |"
                for k in ["train", "validate", "test"]
            ]

            return "\n".join(
                [
                    "| Split      | Students | Percentage |",
                    "|------------|----------|------------|",
                    *rows,
                ]
            )
        else:
            LOGGER.warning(
                "Unable to produce data split table. No splits found in config."
            )
            return f"{card.format.bold('Could not parse data split')}"

    @registry.register("selected_features_ranked_by_shap")
    def selected_features_ranked_by_shap():
        """
        Produces a markdown table of the selected features ranked by average SHAP
        magnitude.
        """
        feature_artifact_path = "selected_features/ranked_selected_features.csv"

        try:
            local_path = utils.download_artifact(
                run_id=card.run_id,
                local_folder=card.assets_folder,
                artifact_path=feature_artifact_path,
            )
            df = pd.read_csv(local_path)

            # Build markdown table
            headers = "| " + " | ".join(df.columns) + " |"
            separator = "| " + " | ".join(["---"] * len(df.columns)) + " |"
            rows = [
                "| " + " | ".join(str(val).replace("\n", "<br>") for val in row) + " |"
                for row in df.values
            ]

            title = f"{card.format.header_level(4)}Selected Features\n"
            subtitle = f"{card.format.header_level(5)}Full List of Selected Features Ranked by Importance\n\n"

            table_markdown = "\n".join([headers, separator] + rows)

            return f"{title}\n{subtitle}{table_markdown}"

        except Exception as e:
            LOGGER.warning(f"Could not load feature importance table: {str(e)}")
            return f"{card.format.bold('Selected Features Ranked by Importance')}\n\nCould not load data."



# Databricks notebook source
# MAGIC %md
# MAGIC # SST Prepare Modeling Dataset: [SCHOOL]
# MAGIC
# MAGIC Second step in the process of transforming raw (PDP) data into actionable, data-driven insights for advisors: featurize the raw, validated data; configure and compute the target variable; perform feature selection; prepare train/test/validation splits; and inspect feature-target associations.
# MAGIC
# MAGIC ### References
# MAGIC
# MAGIC - [Data science product components (Confluence doc)](https://datakind.atlassian.net/wiki/spaces/TT/pages/237862913/Data+science+product+components+the+modeling+process)
# MAGIC - [Databricks Data preparation for classification](https://docs.databricks.com/en/machine-learning/automl/classification-data-prep.html)
# MAGIC - [Databricks runtimes release notes](https://docs.databricks.com/en/release-notes/runtime/index.html)
# MAGIC - [SCHOOL WEBSITE](https://example.com)

# COMMAND ----------

# MAGIC %md
# MAGIC # setup

# COMMAND ----------

# MAGIC %sh python --version

# COMMAND ----------

# install dependencies, of which most/all should come through our 1st-party SST package

# %pip install "student-success-tool == 0.3.7"

# COMMAND ----------

# MAGIC %restart_python

# COMMAND ----------

import functools as ft
import logging
import sys

import matplotlib.pyplot as plt
import pandas as pd  # noqa: F401
import seaborn as sb
from databricks.connect import DatabricksSession

from student_success_tool import configs, dataio, eda, modeling, preprocessing
from student_success_tool.preprocessing import checkpoints, selection, targets

# COMMAND ----------

logging.basicConfig(level=logging.INFO, force=True)
logging.getLogger("py4j").setLevel(logging.WARNING)  # ignore databricks logger

try:
    spark = DatabricksSession.builder.getOrCreate()
except Exception:
    logging.warning("unable to create spark session; are you in a Databricks runtime?")
    pass

# COMMAND ----------

# MAGIC %md
# MAGIC ## import school-specific code

# COMMAND ----------

# MAGIC %load_ext autoreload
# MAGIC %autoreload 2

# COMMAND ----------

# insert our 1st-party (school-specific) code into PATH
if "../" not in sys.path:
    sys.path.insert(1, "../")

# TODO: specify school's subpackage
from pipelines import *  # noqa: F403

# COMMAND ----------

# project configuration should be stored in a config file in TOML format
cfg = dataio.read_config("./config-TEMPLATE.toml", schema=configs.pdp.PDPProjectConfig)
cfg

# COMMAND ----------

# MAGIC %md
# MAGIC # load (+validate) raw data

# COMMAND ----------

# MAGIC %md
# MAGIC **Note:** The specifics of this load+validate process were determined previously in this project's "data assessment" notebook, including which schemas are needed to properly parse and validate the raw data. Also, the paths to the raw files should have been added to the project config. Update the next two cells accordingly.

# COMMAND ----------

raw_course_file_path = cfg.datasets.bronze.raw_course.file_path
df_course = dataio.pdp.read_raw_course_data(
    file_path=raw_course_file_path,
    schema=dataio.schemas.pdp.RawPDPCourseDataSchema,
    dttm_format="%Y%m%d.0",
)
df_course.head()

# COMMAND ----------

raw_cohort_file_path = cfg.datasets.bronze.raw_cohort.file_path
df_cohort = dataio.pdp.read_raw_cohort_data(
    file_path=raw_cohort_file_path, schema=dataio.schemas.pdp.RawPDPCohortDataSchema
)
df_cohort.head()

# COMMAND ----------

# MAGIC %md
# MAGIC # featurize data

# COMMAND ----------

# load featurization params from the project config
feature_params = cfg.preprocessing.features.model_dump()

# COMMAND ----------

df_student_terms = preprocessing.pdp.make_student_term_dataset(
    df_cohort, df_course, **feature_params
)
df_student_terms

# COMMAND ----------

# take a peek at featurized columns -- it's a lot
df_student_terms.columns.tolist()

# COMMAND ----------

# MAGIC %md
# MAGIC # select students and compute targets

# COMMAND ----------

# MAGIC %md
# MAGIC A checkpoint, selection, and target function can be used together to make a labeled dataset for modeling. This involves selecting eligible students, subsetting to just the terms from which predictions are made, and computing target variables. _Which_ function depends primarily on the target to be computed: a failure to earn enough credits within a timeframe since initial enrollment, or a particular mid-way checkpoint (other targets pending). Input parameters will vary depending on the school and the target.
# MAGIC
# MAGIC Here's a concrete example below based on a retention model:
# MAGIC ```python
# MAGIC # school-specific parameters specified in config.toml
# MAGIC [preprocessing.selection]
# MAGIC student_criteria = { enrollment_type = "FIRST-TIME", credential_type_sought_year_1 = ["BACHELOR'S DEGREE", "ASSOCIATE'S Degree"], enrollment_intensity_first_term = ["FULL-TIME", "PART-TIME"] }
# MAGIC intensity_time_limits = { FULL-TIME = [3.0, "year"], PART-TIME = [6.0, "year"] }
# MAGIC
# MAGIC [preprocessing.checkpoint]
# MAGIC name = "end_of_first_term"
# MAGIC type_ = "first"
# MAGIC
# MAGIC [preprocessing.target]
# MAGIC name = "retention"
# MAGIC type_ = "retention"
# MAGIC max_academic_year = "2024"
# MAGIC
# MAGIC # creating checkpoint dataframe
# MAGIC df_ckpt = checkpoints.pdp.first_student_terms_within_cohort(
# MAGIC     df_student_terms,
# MAGIC     sort_cols=cfg.preprocessing.checkpoint.sort_cols,
# MAGIC     include_cols=cfg.preprocessing.checkpoint.include_cols,
# MAGIC )
# MAGIC # selecting students from student criteria
# MAGIC selected_students = selection.pdp.select_students_by_attributes(
# MAGIC     df_student_terms,
# MAGIC     student_id_cols=cfg.student_id_col,
# MAGIC     **cfg.preprocessing.selection.student_criteria,
# MAGIC )
# MAGIC # finding students who meet student criteria in checkpoints
# MAGIC df_labeled = pd.merge(df_ckpt, pd.Series(selected_students.index), how="inner", on=cfg.student_id_col)
# MAGIC
# MAGIC # computing target and adding it to df_labeled
# MAGIC target = targets.pdp.retention.compute_target(
# MAGIC     df_labeled,
# MAGIC     max_academic_year=cfg.preprocessing.target.max_academic_year,
# MAGIC )
# MAGIC df_labeled = pd.merge(df_labeled, target, how="inner", on=cfg.student_id_col)
# MAGIC ```
# MAGIC Check out the `checkpoint`, `selection`, & `targets` modules for options and more info.

# COMMAND ----------

# TODO: choose checkpoint function suitable for school's use case
# parameters should be specified in the config
df_ckpt = checkpoints.pdp.TODO(
    df_student_terms,
    sort_cols=cfg.preprocessing.checkpoint.sort_cols,
    include_cols=cfg.preprocessing.checkpoint.include_cols,
)
df_ckpt

# COMMAND ----------

# select students based on student criteria
selected_students = selection.pdp.select_students_by_attributes(
    df_student_terms,
    student_id_cols=cfg.student_id_col,
    **cfg.preprocessing.selection.student_criteria,
)

# merge checkpoint terms with selected students
df_labeled = pd.merge(
    df_ckpt, pd.Series(selected_students.index), how="inner", on=cfg.student_id_col
)
df_labeled

# COMMAND ----------

# TODO: choose target function suitable for school's use case
# parameters should be specified in the config
target = targets.pdp.TODO.compute_target(df_student_terms, **cfg.preprocessing.target)
df_labeled = pd.merge(df_labeled, target, how="inner", on=cfg.student_id_col)

print(df_labeled[cfg.target_col].value_counts())
print(df_labeled[cfg.target_col].value_counts(normalize=True))

# COMMAND ----------

ax = sb.histplot(
    df_labeled.sort_values("cohort"),
    y="cohort",
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")
plt.show()
ax = sb.histplot(
    df_labeled.sort_values("cohort"),
    y="cohort",
    hue=cfg.target_col,
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")
plt.show()

# COMMAND ----------

# drop unwanted columns and mask values by time
df_preprocessed = preprocessing.pdp.clean_up_labeled_dataset_cols_and_vals(df_labeled)
df_preprocessed.shape

# COMMAND ----------

# MAGIC %md
# MAGIC # splits and sample weights

# COMMAND ----------

try:
    splits = cfg.preprocessing.splits
    split_col = cfg.split_col
    sample_class_weight = cfg.preprocessing.sample_class_weight
    sample_weight_col = cfg.sample_weight_col
except AttributeError:
    splits = {"train": 0.6, "test": 0.2, "validate": 0.2}
    split_col = "split"
    sample_class_weight = "balanced"
    sample_weight_col = "sample_weight"

# COMMAND ----------

if splits:
    df_preprocessed = df_preprocessed.assign(
        **{
            split_col: ft.partial(
                modeling.utils.compute_dataset_splits, seed=cfg.random_state
            )
        }
    )
    print(df_preprocessed[split_col].value_counts(normalize=True))

# COMMAND ----------

if sample_class_weight:
    df_preprocessed = df_preprocessed.assign(
        **{
            sample_weight_col: ft.partial(
                modeling.utils.compute_sample_weights,
                target_col=cfg.target_col,
                class_weight=sample_class_weight,
            )
        }
    )
    print(df_preprocessed[sample_weight_col].value_counts(normalize=True))

# COMMAND ----------

# MAGIC %md
# MAGIC # feature-target relationships

# COMMAND ----------

non_feature_cols = (
    [cfg.student_id_col]
    + (cfg.student_group_cols or [])
    + ([cfg.split_col] if cfg.split_col else [])
    + ([cfg.sample_weight_col] if cfg.sample_weight_col else [])
)

# COMMAND ----------

target_corrs = df_preprocessed.drop(
    columns=non_feature_cols + [cfg.target_col]
).corrwith(df_preprocessed[cfg.target_col], method="spearman", numeric_only=True)
print(target_corrs.sort_values(ascending=False).head(10))
print("...")
print(target_corrs.sort_values(ascending=False, na_position="first").tail(10))

# COMMAND ----------

target_assocs = eda.compute_pairwise_associations(
    df_preprocessed, ref_col=cfg.target_col, exclude_cols=non_feature_cols
)
print(target_assocs.sort_values(by="target", ascending=False).head(10))
print("...")
print(
    target_assocs.sort_values(by="target", ascending=False, na_position="first").tail(
        10
    )
)

# COMMAND ----------

# MAGIC %md
# MAGIC # wrap-up

# COMMAND ----------

# save preprocessed dataset
dataio.write.to_delta_table(
    df_preprocessed, cfg.datasets.silver.preprocessed.table_path, spark_session=spark
)

# COMMAND ----------

# MAGIC %md
# MAGIC - [ ] Submit a PR including this notebook and any changes in project config



import typing as t
from datetime import date

from faker.providers import BaseProvider

from ... import utils


class Provider(BaseProvider):
    def raw_cohort_record(
        self,
        min_cohort_yr: int = 2010,
        max_cohort_yr: t.Optional[int] = None,
        normalize_col_names: bool = False,
        institution_id: t.Optional[str] = None,
        student_guid: t.Optional[str] = None,
    ) -> dict[str, object]:
        # some fields are inputs to others; compute them first, accordingly
        enrollment_type = self.enrollment_type()
        enrollment_intensity_first_term = self.enrollment_intensity_first_term()
        number_of_credits_attempted_year_1 = self.number_of_credits_attempted_year_1()
        number_of_credits_attempted_year_2 = self.number_of_credits_attempted_year_2()
        number_of_credits_attempted_year_3 = self.number_of_credits_attempted_year_3()
        number_of_credits_attempted_year_4 = self.number_of_credits_attempted_year_4()
        _has_enrollment_other_inst: bool = self.generator.random.random() < 0.25

        # TODO: handle other cases, e.g. gateway course attempted/completed/grades
        record = {
            "Student GUID": student_guid
            if student_guid is not None
            else self.student_guid(),
            "Institution ID": institution_id
            if institution_id is not None
            else self.institution_id(),
            "Cohort": self.cohort(min_yr=min_cohort_yr, max_yr=max_cohort_yr),
            "Cohort Term": self.cohort_term(),
            "Student Age": self.student_age(),
            "Enrollment Type": enrollment_type,
            "Enrollment Intensity First Term": enrollment_intensity_first_term,
            "Math Placement": self.math_placement(),
            "English Placement": self.english_placement(),
            "Dual and Summer Enrollment": self.dual_and_summer_enrollment(),
            "Race": self.race(),
            "Ethnicity": self.ethnicity(),
            "Gender": self.gender(),
            "First Gen": self.first_gen(),
            "Pell Status First Year": self.pell_status_first_year(),
            "Attendance Status Term 1": self.attendance_status_term_1(
                enrollment_type, enrollment_intensity_first_term
            ),
            "Credential Type Sought Year 1": self.credential_type_sought_year_1(),
            "Program of Study Term 1": self.program_of_study_term_1(),
            "Program of Study Year 1": self.program_of_study_year_1(),
            "GPA Group Term 1": self.gpa_group_term_1(),
            "GPA Group Year 1": self.gpa_group_year_1(),
            "Number of Credits Attempted Year 1": number_of_credits_attempted_year_1,
            "Number of Credits Earned Year 1": self.number_of_credits_earned_year_1(
                max_value=number_of_credits_attempted_year_1
            ),
            "Number of Credits Attempted Year 2": number_of_credits_attempted_year_2,
            "Number of Credits Earned Year 2": self.number_of_credits_earned_year_2(
                max_value=number_of_credits_attempted_year_2
            ),
            "Number of Credits Attempted Year 3": number_of_credits_attempted_year_3,
            "Number of Credits Earned Year 3": self.number_of_credits_earned_year_3(
                max_value=number_of_credits_attempted_year_3
            ),
            "Number of Credits Attempted Year 4": number_of_credits_attempted_year_4,
            "Number of Credits Earned Year 4": self.number_of_credits_earned_year_4(
                max_value=number_of_credits_attempted_year_4
            ),
            "Gateway Math Status": self.gateway_math_status(),
            "Gateway English Status": self.gateway_english_status(),
            "AttemptedGatewayMathYear1": self.attempted_gateway_math_year_1(),
            "AttemptedGatewayEnglishYear1": self.attempted_gateway_english_year_1(),
            "CompletedGatewayMathYear1": self.completed_gateway_math_year_1(),
            "CompletedGatewayEnglishYear1": self.completed_gateway_english_year_1(),
            "GatewayMathGradeY1": self.gateway_math_grade_y_1(),
            "GatewayEnglishGradeY1": self.gateway_english_grade_y_1(),
            "AttemptedDevMathY1": self.attempted_dev_math_y_1(),
            "AttemptedDevEnglishY1": self.attempted_dev_english_y_1(),
            "CompletedDevMathY1": self.completed_dev_math_y_1(),
            "CompletedDevEnglishY1": self.completed_dev_english_y_1(),
            "Retention": self.retention(),
            "Persistence": self.persistence(),
            "Years to Bachelors at cohort inst.": self.years_to_bachelors_at_cohort_inst(),
            "Years to Associates or Certificate at cohort inst.": self.years_to_associates_or_certificate_at_cohort_inst(),
            "Years to Bachelor at other inst.": self.years_to_bachelor_at_other_inst(),
            "Years to Associates or Certificate at other inst.": self.years_to_associates_or_certificate_at_other_inst(),
            "Years of Last Enrollment at cohort institution": self.years_of_last_enrollment_at_cohort_institution(),
            "Years of Last Enrollment at other institution": self.years_of_last_enrollment_at_other_institution(),
            "Time to Credential": self.time_to_credential(),
            "Reading Placement": self.reading_placement(),
            "Special Program": self.special_program(),
            "NASPA First-Generation": self.naspa_first_generation(),
            "Incarcerated Status": self.incarcerated_status(),
            "Military Status": self.military_status(),
            "Employment Status": self.employment_status(),
            "Disability Status": self.disability_status(),
            "Foreign Language Completion": self.foreign_language_completion(),
            "First Year to Bachelors at cohort inst.": self.first_year_to_bachelors_at_cohort_inst(),
            "First Year to Associates or Certificate at cohort inst.": self.first_year_to_associates_or_certificate_at_cohort_inst(),
            "First Year to Bachelor at other inst.": self.first_year_to_bachelor_at_other_inst(),
            "First Year to Associates or Certificate at other inst.": self.first_year_to_associates_or_certificate_at_other_inst(),
            "Most Recent Bachelors at Other Institution STATE": self.most_recent_bachelors_at_other_institution_state(
                _has_enrollment_other_inst
            ),
            "Most Recent Associates or Certificate at Other Institution STATE": self.most_recent_associates_or_certificate_at_other_institution_state(
                _has_enrollment_other_inst
            ),
            "Most Recent Last Enrollment at Other institution STATE": self.most_recent_last_enrollment_at_other_institution_state(
                _has_enrollment_other_inst
            ),
            "First Bachelors at Other Institution STATE": self.first_bachelors_at_other_institution_state(
                _has_enrollment_other_inst
            ),
            "First Associates or Certificate at Other Institution STATE": self.first_associates_or_certificate_at_other_institution_state(
                _has_enrollment_other_inst
            ),
            "Most Recent Bachelors at Other Institution CARNEGIE": self.most_recent_bachelors_at_other_institution_carnegie(
                _has_enrollment_other_inst
            ),
            "Most Recent Associates or Certificate at Other Institution CARNEGIE": self.most_recent_associates_or_certificate_at_other_institution_carnegie(
                _has_enrollment_other_inst
            ),
            "Most Recent Last Enrollment at Other institution CARNEGIE": self.most_recent_last_enrollment_at_other_institution_carnegie(
                _has_enrollment_other_inst
            ),
            "First Bachelors at Other Institution CARNEGIE": self.first_bachelors_at_other_institution_carnegie(
                _has_enrollment_other_inst
            ),
            "First Associates or Certificate at Other Institution CARNEGIE": self.first_associates_or_certificate_at_other_institution_carnegie(
                _has_enrollment_other_inst
            ),
            "Most Recent Bachelors at Other Institution LOCALE": self.most_recent_bachelors_at_other_institution_locale(
                _has_enrollment_other_inst
            ),
            "Most Recent Associates or Certificate at Other Institution LOCALE": self.most_recent_associates_or_certificate_at_other_institution_locale(
                _has_enrollment_other_inst
            ),
            "Most Recent Last Enrollment at Other institution LOCALE": self.most_recent_last_enrollment_at_other_institution_locale(
                _has_enrollment_other_inst
            ),
            "First Bachelors at Other Institution LOCALE": self.first_bachelors_at_other_institution_locale(
                _has_enrollment_other_inst
            ),
            "First Associates or Certificate at Other Institution LOCALE": self.first_associates_or_certificate_at_other_institution_locale(
                _has_enrollment_other_inst
            ),
        }
        if normalize_col_names:
            record = {
                utils.misc.convert_to_snake_case(key): val
                for key, val in record.items()
            }
        return record

    def student_guid(self) -> str:
        return self.numerify("#####!")

    def institution_id(self) -> str:
        return self.numerify("#####!")

    # Returns a string in the format "YYYY-YY" (e.g. "2010-11"), representing a cohort year,
    # where the first year is the start year (a random date between min_yr and max_yr,
    # or min_yr and now if the current date if not provided). The second year is the first year + 1.
    def cohort(self, min_yr: int = 2010, max_yr: t.Optional[int] = None) -> str:
        _end_date = date(max_yr, 1, 1) if max_yr is not None else "today"
        start_dt: date = self.generator.date_between(
            start_date=date(min_yr, 1, 1), end_date=_end_date
        )
        start_yr = start_dt.year
        end_yr = f"{start_yr + 1}"[2:]
        return f"{start_yr}-{end_yr}"

    def cohort_term(self) -> str:
        return self.random_element(["FALL", "WINTER", "SPRING", "SUMMER"])

    def student_age(self) -> str:
        return self.random_element(["20 AND YOUNGER", ">20 - 24", "OLDER THAN 24"])

    def enrollment_type(self) -> str:
        return self.random_element(["FIRST-TIME", "RE-ADMIT", "TRANSFER-IN"])

    def enrollment_intensity_first_term(self) -> str:
        return self.random_element(["FULL-TIME", "PART-TIME"])

    def math_placement(self) -> str:
        return self.random_element(["C", "N"])

    def english_placement(self) -> str:
        return self.random_element(["C", "N"])

    def dual_and_summer_enrollment(self) -> str:
        return self.random_element(["DE", "SE", "DS"])

    def race(self) -> str:
        return self.random_element(
            [
                "NONRESIDENT ALIEN",
                "AMERICAN INDIAN OR ALASKA NATIVE",
                "ASIAN",
                "BLACK OR AFRICAN AMERICAN",
                "NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER",
                "WHITE",
                "HISPANIC",
                "TWO OR MORE RACES",
                "UNKNOWN",
            ]
        )

    def ethnicity(self) -> str:
        return self.random_element(["H", "N", "UK"])

    def gender(self) -> str:
        return self.random_element(["M", "F", "P", "X", "UK"])

    def first_gen(self) -> str:
        return self.random_element(["P", "C", "A", "B"])

    def pell_status_first_year(self) -> str:
        return self.random_element(["Y", "N"])

    def attendance_status_term_1(
        self,
        enrollment_type: t.Optional[str] = None,
        enrollment_intensity_first_term: t.Optional[str] = None,
    ) -> str:
        if not enrollment_type:
            enrollment_type = self.enrollment_type()
        if not enrollment_intensity_first_term:
            enrollment_intensity_first_term = self.enrollment_intensity_first_term()
        return (
            f"{enrollment_type} {enrollment_intensity_first_term}".title()
            # PDP's string casing is frustratingly inconsistent
            .replace("Admit", "admit")
        )

    def credential_type_sought_year_1(self) -> str:
        return self.random_element(
            [
                "Less than one-year certificate, less than Associate degree",
                "One to two year certificate, less than Associate degree",
                "Two to four year certificate, less than Bachelor's degree",
                "Undergraduate Certificate or Diploma Program",
                "Associate Degree",
                "Bachelor's Degree",
                "Post Baccalaureate Certificate",
                "Master's Degree",
                "Doctoral Degree",
                "First Professional Degree",
                "Graduate/Professional Certificate",
                "Non- Credential Program (Preparatory Coursework/Teach Certification)",
                "Missing",
            ]
        )

    def program_of_study_term_1(self) -> str:
        # TODO: make this six-digit CIP code more realistic
        return self.numerify("##.####")

    def program_of_study_year_1(self) -> str:
        # TODO: make this six-digit CIP code more realistic
        return self.numerify("##.####")

    def _gpa(self) -> float:
        return self.generator.pyfloat(  # type: ignore
            min_value=0.0, max_value=4.0, right_digits=2
        )

    def gpa_group_term_1(self) -> float:
        return self._gpa()

    def gpa_group_year_1(self) -> float:
        return self._gpa()

    def _number_of_credits(
        self, min_value: float = 1.0, max_value: float = 20.0
    ) -> float:
        return self.generator.pyfloat(  # type: ignore
            min_value=min_value,
            max_value=max(max_value, min_value + 1e-3),
            right_digits=1,
        )

    def number_of_credits_attempted_year_1(self) -> float:
        return self._number_of_credits(min_value=1.0)

    def number_of_credits_earned_year_1(self, max_value: float = 20.0) -> float:
        return self._number_of_credits(max_value=max_value)

    def number_of_credits_attempted_year_2(self) -> float:
        return self._number_of_credits()

    def number_of_credits_earned_year_2(self, max_value: float = 20.0) -> float:
        return self._number_of_credits(max_value=max_value)

    def number_of_credits_attempted_year_3(self) -> float:
        return self._number_of_credits()

    def number_of_credits_earned_year_3(self, max_value: float = 20.0) -> float:
        return self._number_of_credits(max_value=max_value)

    def number_of_credits_attempted_year_4(self) -> float:
        return self._number_of_credits()

    def number_of_credits_earned_year_4(self, max_value: float = 20.0) -> float:
        return self._number_of_credits(max_value=max_value)

    def gateway_math_status(self) -> str:
        return self.random_element(["R", "N"])

    def gateway_english_status(self) -> str:
        return self.random_element(["R", "N"])

    def attempted_gateway_math_year_1(self) -> str:
        return self.random_element(["Y", "N"])

    def attempted_gateway_english_year_1(self) -> str:
        return self.random_element(["Y", "N"])

    def completed_gateway_math_year_1(self) -> str:
        return self.random_element(["C", "D", "NA"])

    def completed_gateway_english_year_1(self) -> str:
        return self.random_element(["C", "D", "NA"])

    def _grade(self) -> str:
        # TODO: use weighting for more realistic distribution?
        # find out if grades can ever be given as continuous values
        return self.random_element(
            ["0", "1", "2", "3", "4", "P", "F", "I", "W", "A", "M", "O"]
        )

    def gateway_math_grade_y_1(self) -> str:
        return self._grade()

    def gateway_english_grade_y_1(self) -> str:
        return self._grade()

    def attempted_dev_math_y_1(self) -> str:
        return self.random_element(["Y", "N", "NA"])

    def attempted_dev_english_y_1(self) -> str:
        return self.random_element(["Y", "N", "NA"])

    def completed_dev_math_y_1(self) -> str:
        return self.random_element(["Y", "N", "NA"])

    def completed_dev_english_y_1(self) -> str:
        return self.random_element(["Y", "N", "NA"])

    def retention(self) -> bool:
        return self.generator.pybool()  # type: ignore

    def persistence(self) -> bool:
        return self.generator.pybool()  # type: ignore

    def _years_to_of(self, min_value: int = 0, max_value: int = 7) -> int:
        return self.random_int(min=min_value, max=max_value)

    def years_to_bachelors_at_cohort_inst(self) -> int:
        return self._years_to_of()

    def years_to_associates_or_certificate_at_cohort_inst(self) -> int:
        return self._years_to_of()

    def years_to_bachelor_at_other_inst(
        self, has_enrollment: bool = True
    ) -> t.Optional[int]:
        return self._years_to_of() if has_enrollment else None

    def years_to_associates_or_certificate_at_other_inst(
        self, has_enrollment: bool = True
    ) -> t.Optional[int]:
        return self._years_to_of() if has_enrollment else None

    def years_of_last_enrollment_at_cohort_institution(self) -> int:
        return self._years_to_of()

    def years_of_last_enrollment_at_other_institution(
        self, has_enrollment: bool = True
    ) -> t.Optional[int]:
        return self._years_to_of() if has_enrollment else None

    def time_to_credential(self) -> float:
        return self.generator.pyfloat(  # type: ignore
            min_value=1.0, max_value=20.0, right_digits=0
        )

    def reading_placement(self) -> str:
        return self.random_element(["C", "N"])

    def special_program(self) -> t.Optional[str]:
        # TODO: come up with something here?
        return None

    def naspa_first_generation(self) -> t.Optional[str]:
        return self.random_element(["-1", "0", "1", "2", "3", "4", "5", "6"])

    def incarcerated_status(self) -> t.Optional[str]:
        return self.random_element(["Y", "P", "N"])

    def military_status(self) -> t.Optional[str]:
        return self.random_element(["-1", "0", "1", "2"])

    def employment_status(self) -> t.Optional[str]:
        return self.random_element(["-1", "0", "1", "2", "3", "4"])

    def disability_status(self) -> t.Optional[str]:
        return self.random_element(["Y", "N"])

    def foreign_language_completion(self) -> t.Optional[str]:
        # TODO: come up with something here?
        return None

    def first_year_to_bachelors_at_cohort_inst(self) -> int:
        return self._years_to_of()

    def first_year_to_associates_or_certificate_at_cohort_inst(self) -> int:
        return self._years_to_of()

    def first_year_to_bachelor_at_other_inst(
        self, has_enrollment: bool = True
    ) -> t.Optional[int]:
        return self._years_to_of() if has_enrollment else None

    def first_year_to_associates_or_certificate_at_other_inst(
        self, has_enrollment: bool = True
    ) -> t.Optional[int]:
        return self._years_to_of() if has_enrollment else None

    def _institution_state(self) -> str:
        return self.generator.state_abbr(include_freely_associated_states=False)  # type: ignore

    def most_recent_bachelors_at_other_institution_state(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_state() if has_enrollment else None

    def most_recent_associates_or_certificate_at_other_institution_state(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_state() if has_enrollment else None

    def most_recent_last_enrollment_at_other_institution_state(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_state() if has_enrollment else None

    def first_bachelors_at_other_institution_state(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_state() if has_enrollment else None

    def first_associates_or_certificate_at_other_institution_state(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_state() if has_enrollment else None

    def _institution_carnegie(self) -> str:
        return self.random_element(
            [
                "High traditional",
                "Mixed traditional/nontraditional",
                "High nontraditional",
                "Associate's Dominant",
                "Mixed Baccalaureate/Associate's Colleges",
                "Arts & sciences focus",
                "Diverse fields",
                "M1: Master's Colleges and Universities - Larger programs",
                "M2: Master's Colleges and Universities – Medium programs",
                "M3: Master's Colleges and Universities – Small programs",
                "R1: Doctoral universities – very high research activity",
                "R2: Doctoral universities – high research activity",
                "D/PU: Doctoral/professional universities",
            ]
        )

    def most_recent_bachelors_at_other_institution_carnegie(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_carnegie() if has_enrollment else None

    def most_recent_associates_or_certificate_at_other_institution_carnegie(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_carnegie() if has_enrollment else None

    def most_recent_last_enrollment_at_other_institution_carnegie(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_carnegie() if has_enrollment else None

    def first_bachelors_at_other_institution_carnegie(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_carnegie() if has_enrollment else None

    def first_associates_or_certificate_at_other_institution_carnegie(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_carnegie() if has_enrollment else None

    def _institution_locale(self) -> str:
        return self.random_element(["URBAN", "SUBURB", "TOWN/RURAL"])

    def most_recent_bachelors_at_other_institution_locale(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_locale() if has_enrollment else None

    def most_recent_associates_or_certificate_at_other_institution_locale(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_locale() if has_enrollment else None

    def most_recent_last_enrollment_at_other_institution_locale(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_locale() if has_enrollment else None

    def first_bachelors_at_other_institution_locale(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_locale() if has_enrollment else None

    def first_associates_or_certificate_at_other_institution_locale(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return self._institution_locale() if has_enrollment else None



import os
from contextlib import nullcontext as does_not_raise

import pandas as pd
import pytest
from pandera.errors import SchemaErrors

from student_success_tool import dataio

FIXTURES_PATH = "tests/fixtures"


@pytest.mark.parametrize(
    ["file_name", "schema", "kwargs"],
    [
        ("raw_pdp_course_data.csv", None, None),
        ("raw_pdp_course_data.csv", dataio.schemas.pdp.RawPDPCourseDataSchema, None),
        (
            "raw_pdp_course_data.csv",
            dataio.schemas.pdp.RawPDPCourseDataSchema,
            {"nrows": 1},
        ),
    ],
)
def test_read_raw_pdp_course_data(file_name, schema, kwargs):
    file_path = os.path.join(FIXTURES_PATH, file_name)
    result = dataio.pdp.read_raw_course_data(
        file_path=file_path, schema=schema, dttm_format="%Y%m%d", **(kwargs or {})
    )
    assert isinstance(result, pd.DataFrame)
    assert not result.empty


@pytest.mark.parametrize(
    ["file_name", "schema", "converter_func", "exp_ctx"],
    [
        (
            "raw_pdp_course_data_invalid.csv",
            dataio.schemas.pdp.RawPDPCourseDataSchema,
            None,
            pytest.raises(SchemaErrors),
        ),
        (
            "raw_pdp_course_data_invalid.csv",
            dataio.schemas.pdp.RawPDPCourseDataSchema,
            lambda df: df.drop_duplicates(subset=["institution_id", "student_guid"]),
            does_not_raise(),
        ),
    ],
)
def test_read_raw_pdp_course_data_convert(file_name, schema, converter_func, exp_ctx):
    file_path = os.path.join(FIXTURES_PATH, file_name)
    with exp_ctx:
        result = dataio.pdp.read_raw_course_data(
            file_path=file_path,
            schema=schema,
            dttm_format="%Y%m%d",
            converter_func=converter_func,
        )
        assert isinstance(result, pd.DataFrame)
        assert not result.empty


@pytest.mark.parametrize(
    ["file_name", "schema", "kwargs"],
    [
        ("raw_pdp_cohort_data.csv", None, None),
        ("raw_pdp_cohort_data.csv", dataio.schemas.pdp.RawPDPCohortDataSchema, None),
        # Yes and No replace 1 and 0.
        ("raw_pdp_cohort_data_ys.csv", dataio.schemas.pdp.RawPDPCohortDataSchema, None),
        (
            "raw_pdp_cohort_data.csv",
            dataio.schemas.pdp.RawPDPCohortDataSchema,
            {"nrows": 1},
        ),
    ],
)
def test_read_raw_pdp_cohort_data(file_name, schema, kwargs):
    file_path = os.path.join(FIXTURES_PATH, file_name)
    result = dataio.pdp.read_raw_cohort_data(
        file_path=file_path, schema=schema, **(kwargs or {})
    )
    assert isinstance(result, pd.DataFrame)
    assert not result.empty


@pytest.mark.parametrize(
    ["df", "col", "fmt", "exp"],
    [
        (
            pd.DataFrame({"dttm": ["2024-01-01", "2024-01-02"]}, dtype="string"),
            "dttm",
            "%Y-%m-%d",
            pd.Series([pd.Timestamp(2024, 1, 1), pd.Timestamp(2024, 1, 2)]),
        ),
        (
            pd.DataFrame({"foo": ["20240101.0", "20240102.0"]}, dtype="string"),
            "foo",
            "%Y%m%d.0",
            pd.Series([pd.Timestamp(2024, 1, 1), pd.Timestamp(2024, 1, 2)]),
        ),
    ],
)
def test_parse_dttm_values(df, col, fmt, exp):
    obs = dataio.pdp.raw_data._parse_dttm_values(df, col=col, fmt=fmt)
    assert obs.equals(exp)


@pytest.mark.parametrize(
    ["df", "col", "exp"],
    [
        (
            pd.DataFrame({"col1": ["a", "b", "c"]}, dtype="string"),
            "col1",
            pd.Series(["A", "B", "C"], dtype="string"),
        ),
        (
            pd.DataFrame({"col2": ["aBc", "BcD", "cDe"]}, dtype="string"),
            "col2",
            pd.Series(["ABC", "BCD", "CDE"], dtype="string"),
        ),
        (
            pd.DataFrame({"col": []}, dtype="string"),
            "col",
            pd.Series([], dtype="string"),
        ),
    ],
)
def test_uppercase_string_values(df, col, exp):
    obs = dataio.pdp.raw_data._uppercase_string_values(df, col=col)
    assert obs.equals(exp)


@pytest.mark.parametrize(
    ["df", "col", "to_replace", "exp"],
    [
        (
            pd.DataFrame({"gpa1": ["4.0", "2.5", "UK"]}),
            "gpa1",
            "UK",
            pd.Series(["4.0", "2.5", None]),
        ),
        (
            pd.DataFrame({"col": ["-1.0", "1", "2"]}),
            "col",
            "-1.0",
            pd.Series([None, "1", "2"]),
        ),
    ],
)
def test_replace_values_with_null(df, col, to_replace, exp):
    obs = dataio.pdp.raw_data._replace_values_with_null(
        df, col=col, to_replace=to_replace
    )
    assert obs.equals(exp)


@pytest.mark.parametrize(
    ["df", "col", "exp"],
    [
        (
            pd.DataFrame({"col1": ["1", "0", "1"]}, dtype="string"),
            "col1",
            pd.Series([True, False, True], dtype="boolean"),
        ),
        (
            pd.DataFrame({"col2": ["1", "0", None]}, dtype="string"),
            "col2",
            pd.Series([True, False, None], dtype="boolean"),
        ),
        (
            pd.DataFrame({"col3": ["True", "False"]}, dtype="string"),
            "col3",
            pd.Series([True, False], dtype="boolean"),
        ),
    ],
)
def test_cast_to_bool_via_int(df, col, exp):
    obs = dataio.pdp.raw_data._cast_to_bool_via_int(df, col=col)
    assert obs.equals(exp)



from .. import (
    bias_sections as base_bias_sections,
)


def register_bias_sections(card, registry):
    base_bias_sections.register_bias_sections(card, registry)

    @registry.register("bias_groups_section")
    def bias_groups_section():
        """
        Returns bias groups for PDP. These groups will be static across
        all institutions in PDP.
        """
        intro = f"{card.format.indent_level(1)}- Our assessment for FNR Parity was conducted across the following student groups.\n"
        groups = [
            "Ethnicity",
            "First Generation Status",
            "Gender",
            "Race",
            "Student Age",
        ]
        nested = [f"{card.format.indent_level(2)}- {group}\n" for group in groups]
        return intro + "".join(nested)



from collections.abc import Iterable

import pytest

from student_success_tool import utils


@pytest.mark.parametrize(
    ["val", "exp"],
    [
        ([1, 2, 3], [1, 2, 3]),
        ((1, 2, 3), [1, 2, 3]),
        ({"a": 1, "b": 2}, ["a", "b"]),
        ("abc", ["abc"]),
        (1, [1]),
        (None, [None]),
    ],
)
def test_to_list(val, exp):
    obs = utils.types.to_list(val)
    assert obs == exp


@pytest.mark.parametrize(
    ["value", "exp"],
    [
        ([1, 2, 3], True),
        ({"a", "b", "c"}, True),
        ((True, False, True), True),
        ("string", False),
        (b"bytes", False),
    ],
)
def test_is_collection_but_not_string(value, exp):
    obs = utils.types.is_collection_but_not_string(value)
    assert isinstance(obs, bool)
    assert obs == exp


@pytest.mark.parametrize(
    ["eles", "exp"],
    [
        ([2, 1, 2, 2, 1, 3], [2, 1, 3]),
        (("a", "c", "b", "b", "c", "a"), ["a", "c", "b"]),
    ],
)
def test_unique_elements_in_order(eles, exp):
    obs = utils.misc.unique_elements_in_order(eles)
    assert isinstance(obs, Iterable)
    assert list(obs) == exp


@pytest.mark.parametrize(
    ["val", "exp"],
    [
        ("Student GUID", "student_guid"),
        ("Credential Type Sought Year 1", "credential_type_sought_year_1"),
        ("Years to Bachelors at cohort inst.", "years_to_bachelors_at_cohort_inst"),
        ("Enrolled at Other Institution(s)", "enrolled_at_other_institution_s"),
    ],
)
def test_convert_to_snake_case(val, exp):
    obs = utils.misc.convert_to_snake_case(val)
    assert obs == exp



import itertools
import logging
import typing as t

import numpy as np
import pandas as pd
import scipy.stats as ss

from . import utils

LOGGER = logging.getLogger(__name__)


def assess_unique_values(data: pd.DataFrame, cols: str | list[str]) -> dict[str, int]:
    """
    Assess unique values in ``data`` given by the combination of columns in ``cols`` ,
    including counts of nunique, duplicates, and nulls.

    Args:
        data
        cols
    """
    unique_data = data.loc[:, utils.types.to_list(cols)]
    is_duplicated = unique_data.duplicated()
    return {
        "num_uniques": is_duplicated.eq(False).sum(),
        "num_dupes": is_duplicated.sum(),
        "num_with_null_values": unique_data.isna().sum(axis="columns").gt(0).sum(),
    }


def compute_summary_stats(
    data: pd.DataFrame,
    *,
    include: t.Optional[str | list[str]] = None,
    exclude: t.Optional[str | list[str]] = None,
    percentiles: t.Optional[list[float]] = None,
) -> pd.DataFrame:
    """
    Compute summary stats for columns in ``data`` matching one or multiple dtypes
    using standard :meth:`pd.DataFrame.describe()` , supplemented with null count/pct.

    Args:
        data
        include: One or multiple dtypes whose columns will be included in result.
        exclude: One or multiple dtypes whose columns will be excluded from result.
        percentiles: Percentiles to include in result, given as floats between 0 and 1.

    References:
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.select_dtypes.html
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.describe.html
    """
    num_rows = data.shape[0]
    include = utils.types.to_list(include) if include is not None else None
    exclude = utils.types.to_list(exclude) if exclude is not None else None
    data_selected = data.select_dtypes(include=include, exclude=exclude)  # type: ignore
    data_described = data_selected.describe(percentiles=percentiles).T.assign(
        null_count=data_selected.isna().sum(),
        null_pct=lambda df: (100 * df["null_count"] / num_rows).round(decimals=1),
    )
    return data_described


def compute_group_counts_pcts(
    data: pd.DataFrame,
    cols: str | list[str],
    *,
    sort: bool = True,
    ascending: bool = False,
    dropna: bool = True,
) -> pd.DataFrame:
    """
    Compute value counts and percent totals in ``data`` for groups defined by ``cols`` .

    Args:
        data
        cols
        sort
        ascending
        dropna

    References:
        - https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.value_counts.html
    """
    return (
        # data.value_counts(cols, dropna = False) for some reason still drops NaNs. We use
        # data[cols].value_counts() to avoid this error
        data[cols]
        .value_counts(sort=sort, ascending=ascending, dropna=dropna)
        .to_frame(name="count")
        .assign(
            pct=lambda df: (100 * df["count"] / df["count"].sum()).round(decimals=1)
        )
    )


def compute_crosstabs(
    data: pd.DataFrame,
    index_cols: str | list[str],
    column_cols: str | list[str],
    value_col: t.Optional[str] = None,
    aggfunc: t.Optional[t.Callable] = None,
    margins: bool = True,
    normalize: bool | t.Literal["all", "index", "columns"] = False,
) -> pd.DataFrame:
    """
    Args:
        data
        index_cols
        column_cols
        value_col
        aggfunc
        margins
        normalize

    References:
        - https://pandas.pydata.org/docs/reference/api/pandas.crosstab.html
    """
    index = (
        data[index_cols]
        if isinstance(index_cols, str)
        else [data[col] for col in index_cols]
    )
    columns = (
        data[column_cols]
        if isinstance(column_cols, str)
        else [data[col] for col in column_cols]
    )
    values = None if value_col is None else data[value_col]
    ct = pd.crosstab(
        index,
        columns,
        values=values,  # type: ignore
        aggfunc=aggfunc,  # type: ignore
        margins=margins,
        normalize=normalize,
    )
    assert isinstance(ct, pd.DataFrame)  # type guard
    if normalize is not False:
        ct = ct.round(decimals=3)
    return ct


def compute_pairwise_associations(
    df: pd.DataFrame,
    *,
    ref_col: t.Optional[str] = None,
    exclude_cols: t.Optional[str | list[str]] = None,
) -> pd.DataFrame:
    """
    Compute pairwise associations between all columns and each other or, instead,
    all columns and a specified reference column.

    Per-pair association metrics depend on the data types of each:

        - nominal-nominal => Cramer's V
        - numeric-numeric => Spearman rank correlation
        - nominal-numeric => Correlation ratio

    Args:
        df
        ref_col: Reference column against which associations are to be computed.
            If None, all pairwise associations are computed.
        exclude_cols: One or multiple columns to exclude from computing associations;
            for example, if values are unique identifiers or all a single value,
            making their associations irrelevant.

    References:
        - https://en.wikipedia.org/wiki/Cram%C3%A9r%27s_V
        - https://en.wikipedia.org/wiki/Correlation_ratio
    """
    # cast datetime columns to numeric, boolean to string
    df = df.assign(
        **{
            col: pd.to_numeric(df[col])
            for col in df.select_dtypes(include="datetime").columns
        }
        | {
            col: df[col].astype("string")
            for col in df.select_dtypes(include="boolean").columns
        }
    )
    # identify and organize columns in df
    if exclude_cols:
        df = df.drop(columns=exclude_cols)
    cols = df.columns.tolist()
    nominal_cols = set(
        df.select_dtypes(include=["category", "string", "boolean"]).columns.tolist()
    )
    numeric_cols = set(df.select_dtypes(include="number").columns.tolist())
    single_value_cols = _get_single_value_columns(df)
    # store col-col association values
    ref_cols = cols if ref_col is None else [ref_col]
    df_assoc = pd.DataFrame(index=cols, columns=ref_cols, dtype="Float32")
    for col1, col2 in itertools.product(cols, ref_cols):
        if not pd.isna(df_assoc.at[col1, col2]):
            continue

        is_symmetric = False
        if col1 == col2:  # self-association
            assoc = 1.0
        elif col1 in single_value_cols or col2 in single_value_cols:  # n/a
            assoc = None
        elif col1 in nominal_cols and col2 in nominal_cols:  # nom-nom
            assoc = _cramers_v(df[col1], df[col2])
            is_symmetric = True
        elif (col1 in nominal_cols and col2 in numeric_cols) or (
            col1 in numeric_cols and col2 in nominal_cols
        ):  # nom-num
            assoc = _correlation_ratio(df[col1], df[col2])
        elif col1 in numeric_cols and col2 in numeric_cols:  # num-num
            assoc = df[col1].corr(df[col2], method="spearman")
            is_symmetric = True
        else:
            LOGGER.warning(
                "'%s' and/or '%s' columns' dtypes (%s and/or %s) aren't supported "
                "for association computation; skipping ...",
                col1,
                col2,
                df[col1].dtype,
                df[col2].dtype,
            )
            assoc = None

        df_assoc.loc[col1, col2] = assoc
        if is_symmetric and len(ref_cols) > 1:
            df_assoc.loc[col2, col1] = assoc
        if assoc is not None:
            LOGGER.debug("%s – %s association = %s", col1, col2, assoc)
    return df_assoc


def _get_single_value_columns(df: pd.DataFrame) -> set[str]:
    sv_cols = []
    for col in df.columns:
        try:
            nunique = df[col].nunique()
        except TypeError:  # womp
            continue
        if nunique == 1:
            sv_cols.append(col)
    return set(sv_cols)


def _cramers_v(s1: pd.Series, s2: pd.Series) -> float | None:
    """
    Compute Cramer's V statistic for nominal-nominal association,
    which is symmetric -- i.e. V(x, y) == V(y, x).

    References:
        - https://en.wikipedia.org/wiki/Cram%C3%A9r%27s_V

    See Also:
        - :func:`scipy.stats.contingency.association()`
    """
    if not pd.api.types.is_string_dtype(s1) or not pd.api.types.is_string_dtype(s2):
        raise ValueError()

    s1, s2 = _drop_incomplete_pairs(s1, s2)
    if s1.empty or s2.empty:
        return None

    confusion_matrix = pd.crosstab(s1, s2)
    correction = False if confusion_matrix.shape[0] == 2 else True
    try:
        result = ss.contingency.association(
            confusion_matrix, method="cramer", correction=correction
        )
        assert isinstance(result, float)
        return result
    except ValueError:
        return None


def _correlation_ratio(s1: pd.Series, s2: pd.Series) -> float | None:
    """
    Compute the Correlation Ratio for nominal-numeric association.

    References:
        - https://en.wikipedia.org/wiki/Correlation_ratio

    Note:
        ``s1`` and ``s2`` are automatically detected as being categorical or numeric,
        and handled correspondingly in the calculations.
    """
    if pd.api.types.is_string_dtype(s1) and pd.api.types.is_numeric_dtype(s2):
        categories = s1
        measurements = s2
    elif pd.api.types.is_numeric_dtype(s1) and pd.api.types.is_string_dtype(s2):
        categories = s2
        measurements = s1
    else:
        raise ValueError()

    categories, measurements = _drop_incomplete_pairs(categories, measurements)
    if categories.empty or measurements.empty:
        return None

    fcat, _ = pd.factorize(categories)
    cat_num = np.max(fcat) + 1
    y_avg_array = np.zeros(cat_num)
    n_array = np.zeros(cat_num)
    for i in range(0, cat_num):
        cat_measures = measurements[np.argwhere(fcat == i).flatten()]
        n_array[i] = len(cat_measures)
        y_avg_array[i] = np.average(cat_measures)
    y_total_avg = np.sum(np.multiply(y_avg_array, n_array)) / np.sum(n_array)
    numerator = np.sum(
        np.multiply(n_array, np.power(np.subtract(y_avg_array, y_total_avg), 2))
    )
    denominator = np.sum(np.power(np.subtract(measurements, y_total_avg), 2))
    if numerator == 0:
        eta = 0.0
    else:
        eta = np.sqrt(numerator / denominator)
    return eta


def _drop_incomplete_pairs(s1: pd.Series, s2: pd.Series) -> tuple[pd.Series, pd.Series]:
    df = pd.DataFrame({"s1": s1, "s2": s2})
    df = df.dropna(axis="index", how="any", ignore_index=True)  # type: ignore
    return (df["s1"], df["s2"])



import pandas as pd
import pytest

from student_success_tool.preprocessing.targets.pdp import graduation


@pytest.mark.parametrize(
    [
        "df",
        "intensity_time_limits",
        "num_terms_in_year",
        "max_term_rank",
        "student_id_cols",
        "exp",
    ],
    [
        # base case: all students labelable, one pos and one neg
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "02", "02"],
                    "enrollment_intensity": ["PT", "FT", "FT", "PT", "PT"],
                    "years_to_degree": [4, 4, 4, 5, 5],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "term_rank": [1, 2, 3, 1, 2],
                    "term_is_pre_cohort": [True, False, False, False, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [3, "year"], "PT": [6, "year"]},
            2,
            15,
            "student_id",
            pd.Series(
                data=[True, False],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                dtype="boolean",
                name="target",
            ),
        ),
        # lower max target term so part-time student isn't labelable
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "02", "02"],
                    "enrollment_intensity": ["PT", "FT", "FT", "PT", "PT"],
                    "years_to_degree": [4, 4, 4, 5, 5],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "term_rank": [1, 2, 3, 1, 2],
                    "term_is_pre_cohort": [True, False, False, False, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [3, "year"], "PT": [6, "year"]},
            2,
            12,
            "student_id",
            pd.Series(
                data=[True],
                index=pd.Index(["01"], dtype="string", name="student_id"),
                dtype="boolean",
                name="target",
            ),
        ),
        # one time limit for all enrollment intensities
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "02", "02"],
                    "enrollment_intensity": ["PT", "FT", "FT", "PT", "PT"],
                    "years_to_degree": [4, 4, 4, 5, 5],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "term_rank": [1, 2, 3, 1, 2],
                    "term_is_pre_cohort": [True, False, False, False, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"*": [3, "year"]},
            2,
            15,
            "student_id",
            pd.Series(
                data=[True, True],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                dtype="boolean",
                name="target",
            ),
        ),
        # reduce full-time / increase part-time students' years-to-degree
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "02", "02"],
                    "enrollment_intensity": ["PT", "FT", "FT", "PT", "PT"],
                    "years_to_degree": [3, 3, 3, 7, 7],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "term_rank": [1, 2, 3, 1, 2],
                    "term_is_pre_cohort": [True, False, False, False, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [3, "year"], "PT": [6, "year"]},
            2,
            15,
            "student_id",
            pd.Series(
                data=[False, True],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                dtype="boolean",
                name="target",
            ),
        ),
        # pathological case: years-to-degree varies across student terms
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "02", "02"],
                    "enrollment_intensity": ["PT", "FT", "FT", "PT", "PT"],
                    "years_to_degree": [4, 5, 5, pd.NA, 5],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "term_rank": [1, 2, 3, 1, 2],
                    "term_is_pre_cohort": [True, False, False, False, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [3, "year"], "PT": [6, "year"]},
            2,
            15,
            "student_id",
            pd.Series(
                data=[True, False],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                dtype="boolean",
                name="target",
            ),
        ),
    ],
)
def test_compute_target(
    df, intensity_time_limits, num_terms_in_year, max_term_rank, student_id_cols, exp
):
    obs = graduation.compute_target(
        df,
        intensity_time_limits=intensity_time_limits,
        num_terms_in_year=num_terms_in_year,
        max_term_rank=max_term_rank,
        student_id_cols=student_id_cols,
        enrollment_intensity_col="enrollment_intensity",
        years_to_degree_col="years_to_degree",
        enrollment_year_col="enrollment_year",
    )
    assert isinstance(obs, pd.Series)
    print("obs:", obs)
    print("exp:", exp)
    assert pd.testing.assert_series_equal(obs, exp) is None



"""Utility functions for sending emails in Databricks."""

import smtplib

from email.message import EmailMessage

SMTP_SERVER = "smtp.mandrillapp.com"
# TODO: switch port?
SMTP_PORT = 587  # or 465 for SSL
COMPLETION_SUCCESS_SUBJECT = "Student Success Tool: Inference Results Available."
COMPLETION_SUCCESS_MESSAGE = """\
    Hello!

    Your Datakind Student Success Tool inference run has successfully completed and the results are ready for viewing on the Website.
    """

INFERENCE_KICKOFF_SUBJECT = "Student Success Tool: Inference Run In Progress."
INFERENCE_KICKOFF_MESSAGE = """\
    Hello!

    Your Datakind Student Success Tool inference run has been triggered. Once results have been checked and are ready for viewing, you'll receive another email.
    """


def send_email(
    sender_email: str,
    receiver_email_list: list[str],
    cc_email_list: list[str],
    subject: str,
    body: str,
    mandrill_username: str,
    mandrill_password: str,
) -> None:
    """Send email.

    Args:
      sender_email: Email of sender.
      receiver_email_list: List of emails to send to.
      cc_email_list: List of emails to CC.
      subject: Subject of email.
      body: Body of email.
      mandrill_username: Mandrill username for the SMTP server.
      mandrill_password: Mandrill password for the SMTP server.

    Returns:
      Nothing.
    """
    # Create the email message
    msg = EmailMessage()
    msg["Subject"] = subject
    msg["From"] = sender_email
    msg["To"] = receiver_email_list
    msg["Cc"] = cc_email_list
    msg.set_content(body)
    try:
        server = smtplib.SMTP(SMTP_SERVER, SMTP_PORT)
        server.ehlo()
        server.starttls()
        server.login(mandrill_username, mandrill_password)
        server.send_message(msg)
        server.quit()
    except Exception as e:
        raise e


def send_inference_completion_email(
    sender_email: str,
    receiver_email_list: list[str],
    cc_email_list: list[str],
    username: str,
    password: str,
) -> None:
    """Send email with completion of inference run message.

    Args:
      sender_email: Email of sender.
      receiver_email_list: List of emails to send to.
      cc_email_list: List of emails to CC.
      username: Mandrill username for the SMTP server.
      password: Mandrill password for the SMTP server.

    Returns:
      Nothing.
    """
    send_email(
        sender_email,
        receiver_email_list,
        cc_email_list,
        COMPLETION_SUCCESS_SUBJECT,
        COMPLETION_SUCCESS_MESSAGE,
        username,
        password,
    )


def send_inference_kickoff_email(
    sender_email: str,
    receiver_email_list: list[str],
    cc_email_list: list[str],
    username: str,
    password: str,
) -> None:
    """Send email with kickoff of inference run message.

    Args:
      sender_email: Email of sender.
      receiver_email_list: List of emails to send to.
      cc_email_list: List of emails to CC.
      username: Mandrill username for the SMTP server.
      password: Mandrill password for the SMTP server.

    Returns:
      Nothing.
    """
    send_email(
        sender_email,
        receiver_email_list,
        cc_email_list,
        INFERENCE_KICKOFF_SUBJECT,
        INFERENCE_KICKOFF_MESSAGE,
        username,
        password,
    )



import typing as t
from mlflow.tracking import MlflowClient

from ...configs.pdp import PDPProjectConfig
from .base import ModelCard
from ..sections.pdp import register_sections as register_pdp_sections


class PDPModelCard(ModelCard[PDPProjectConfig]):
    def __init__(
        self,
        config: PDPProjectConfig,
        catalog: str,
        model_name: str,
        assets_path: t.Optional[str] = None,
        mlflow_client: t.Optional[MlflowClient] = None,
    ):
        """
        Initializes PDP model card by enforcing a PDP project config.
        Otherwise, this class inherits and is functionally the same as the
        base ModelCard class.
        """
        if not isinstance(config, PDPProjectConfig):  # type guard
            raise TypeError("Expected config to be of type PDPProjectConfig")

        super().__init__(config, catalog, model_name, assets_path, mlflow_client)

    def _register_sections(self):
        """
        Register PDP-specific sections.
        """
        # Clearing registry for overrides
        self.section_registry.clear()

        # Register PDP-specific sections
        register_pdp_sections(self, self.section_registry)



"""Task for publishing data that has been approved and deleting the unapproved outputs for that job.

NOTE: this job does NOT move the existing bucket file from unapproved/ to approved/. It instead writes
the current files in the Databricks volume to approved/ and deletes the unapproved/ output. Which means,
if the actual file in the Databricks volume gets updated, the webapp will pick up the updates. And the
old output will be DELETED.
"""

import argparse
import logging
from google.cloud import storage

from student_success_tool.utils.gcs import publish_inference_output_files
from student_success_tool.utils.emails import send_email
from databricks.sdk import WorkspaceClient
from email.headerregistry import Address

APPROVAL_SUBJECT = "Student Success Tool: Inference Output Reviewed By Datakind."
APPROVAL_MESSAGE = """\
    Hello!

    Your Datakind Student Success Tool inference output has been reviewed by Datakind.
    """


def main():
    """Main function."""
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--db_workspace", required=True, help="Databricks workspace of the task."
    )
    parser.add_argument(
        "--institution_name",
        required=True,
        help="IThe Databricks institution name whose output we want to publish",
    )
    parser.add_argument(
        "--external_bucket_name",
        required=True,
        help="Name of the bucket we want to publish results to (in format [env]_[inst_id]).",
    )
    parser.add_argument(
        "--sst_job_id",
        required=True,
        help="The job run id of the current pipeline execution.",
    )
    parser.add_argument(
        "--email_recipient",
        required=True,
        help="User's email who triggered the inference run.",
    )
    args = parser.parse_args()
    logging.info("Publishing files as approved to GCP bucket")
    publish_inference_output_files(
        args.db_workspace,
        args.institution_name,
        args.external_bucket_name,
        args.sst_job_id,
        True,  # Set approved = True since this task runs after human approval.
    )
    logging.info("Delete unapproved files for this job from GCP bucket.")
    w = WorkspaceClient()
    unapproved_path = f"unapproved/{args.sst_job_id}/"
    storage_client = storage.Client()
    bucket = storage_client.bucket(args.external_bucket_name)
    blobs = bucket.list_blobs(prefix=unapproved_path)
    for blob in blobs:
        blob.delete()
    username = w.dbutils.secrets.get(scope="sst", key="MANDRILL_USERNAME")
    sender_email = Address("Datakind Info", "help", "datakind.org")
    cc_email_list = ["education@datakind.org"]
    password = w.dbutils.secrets.get(scope="sst", key="MANDRILL_PASSWORD")
    logging.info("Sending email notification")
    send_email(
        sender_email,
        [args.email_recipient],
        cc_email_list,
        APPROVAL_SUBJECT,
        APPROVAL_MESSAGE,
        username,
        password,
    )


if __name__ == "__main__":
    main()



import functools as ft
import logging
import typing as t

import numpy as np
import pandas as pd

from . import constants, shared

LOGGER = logging.getLogger(__name__)


def aggregate_from_course_level_features(
    df: pd.DataFrame,
    *,
    student_term_id_cols: list[str],
    min_passing_grade: float = constants.DEFAULT_MIN_PASSING_GRADE,
    key_course_subject_areas: t.Optional[list[str]] = None,
    key_course_ids: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    Aggregate course-level features up to student-term-level features
    by grouping on ``student_term_id_cols`` , then aggregating columns' values
    by specified functions or as dummy columns whose values are, in turn, summed.

    Args:
        df
        student_term_id_cols: Columns that uniquely identify student-terms,
            used to group rows in ``df`` and merge features back in.
        min_passing_grade: Minimum numeric grade considered by institution as "passing".
            Default value is 1.0, i.e. a "D" grade or better.
        key_course_subject_areas: List of course subject areas that are particularly
            relevant ("key") to the institution, such that features are computed to
            measure the number of courses falling within them per student-term.
        key_course_ids

    See Also:
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.get_dummies.html
        - https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#built-in-aggregation-methods

    Notes:
        Rows for which any value in ``student_term_id_cols`` is null are dropped
        and features aren't computed! This is because such a group is "undefined",
        so we can't know if the resulting features are correct.
    """
    LOGGER.info("aggregating course-level data to student-term-level features ...")
    df_grped = df.groupby(by=student_term_id_cols, observed=True, as_index=False)
    # pass through useful metadata and term features as-is
    # assumed to have the same values for every row per group
    df_passthrough = df_grped.agg(
        institution_id=("institution_id", "first"),
        academic_year=("academic_year", "first"),
        academic_term=("academic_term", "first"),
        term_start_dt=("term_start_dt", "first"),
        term_rank=("term_rank", "first"),
        term_rank_core=("term_rank_core", "first"),
        term_rank_noncore=("term_rank_noncore", "first"),
        term_is_core=("term_is_core", "first"),
        term_is_noncore=("term_is_noncore", "first"),
        term_in_peak_covid=("term_in_peak_covid", "first"),
        term_program_of_study=("term_program_of_study", "first"),
    )
    # various aggregations, with an eye toward cumulative features downstream
    df_aggs = df_grped.agg(
        num_courses=num_courses_col_agg(),
        num_courses_passed=num_courses_passed_col_agg(),
        num_courses_completed=num_courses_completed_col_agg(),
        num_credits_attempted=num_credits_attempted_col_agg(),
        num_credits_earned=num_credits_earned_col_agg(),
        course_ids=course_ids_col_agg(),
        course_subjects=course_subjects_col_agg(),
        course_subject_areas=course_subject_areas_col_agg(),
        course_id_nunique=course_id_nunique_col_agg(),
        course_subject_nunique=course_subject_nunique_col_agg(),
        course_subject_area_nunique=course_subject_area_nunique_col_agg(),
        course_level_mean=course_level_mean_col_agg(),
        course_level_std=course_level_std_col_agg(),
        course_grade_numeric_mean=course_grade_numeric_mean_col_agg(),
        course_grade_numeric_std=course_grade_numeric_std_col_agg(),
        section_num_students_enrolled_mean=section_num_students_enrolled_mean_col_agg(),
        section_num_students_enrolled_std=section_num_students_enrolled_std_col_agg(),
        sections_num_students_enrolled=sections_num_students_enrolled_col_agg(),
        sections_num_students_passed=sections_num_students_passed_col_agg(),
        sections_num_students_completed=sections_num_students_completed_col_agg(),
    )
    df_dummies = sum_dummy_cols_by_group(
        df,
        grp_cols=student_term_id_cols,
        agg_cols=[
            "course_type",
            "delivery_method",
            "math_or_english_gateway",
            "co_requisite_course",
            "course_instructor_employment_status",
            "course_instructor_rank",
            "course_level",
            "course_grade",
        ],
    )

    agg_col_vals: list[tuple[str, t.Any | list[t.Any]]] = [
        ("core_course", "Y"),
        ("course_type", ["CC", "CD"]),
        ("course_level", [0, 1]),
        ("enrolled_at_other_institution_s", "Y"),
    ]
    if key_course_subject_areas is not None:
        agg_col_vals.extend(
            ("course_subject_area", kcsa) for kcsa in key_course_subject_areas
        )
    if key_course_ids is not None:
        agg_col_vals.extend(("course_id", kc) for kc in key_course_ids)
    df_val_equals = sum_val_equal_cols_by_group(
        df, grp_cols=student_term_id_cols, agg_col_vals=agg_col_vals
    )
    df_dummy_equals = equal_cols_by_group(
        df=df_val_equals, grp_cols=student_term_id_cols
    )
    df_grade_aggs = multicol_grade_aggs_by_group(
        df, min_passing_grade=min_passing_grade, grp_cols=student_term_id_cols
    )
    return shared.merge_many_dataframes(
        [
            df_passthrough,
            df_aggs,
            df_val_equals,
            df_dummy_equals,
            df_dummies,
            df_grade_aggs,
        ],
        on=student_term_id_cols,
    )


def add_features(
    df: pd.DataFrame,
    *,
    min_num_credits_full_time: float = constants.DEFAULT_MIN_NUM_CREDITS_FULL_TIME,
) -> pd.DataFrame:
    """
    Compute various student-term-level features from aggregated course-level features
    joined to student-level features.

    Args:
        df
        min_num_credits_full_time: Minimum number of credits *attempted* per term
            for a student's enrollment intensity to be considered "full-time".
            Default value is 12.0.

    See Also:
        - :func:`aggregate_from_course_level_features()`
    """
    LOGGER.info("adding student-term features ...")
    nc_prefix = constants.NUM_COURSE_FEATURE_COL_PREFIX
    fc_prefix = constants.FRAC_COURSE_FEATURE_COL_PREFIX
    _num_course_cols = (
        [col for col in df.columns if col.startswith(f"{nc_prefix}_")]
        +
        # also include num-course cols to be added below
        [
            "num_courses_in_program_of_study_area_term_1",
            "num_courses_in_program_of_study_area_year_1",
            "num_courses_in_term_program_of_study_area",
        ]
    )
    num_frac_courses_cols = [
        (col, col.replace(f"{nc_prefix}_", f"{fc_prefix}_")) for col in _num_course_cols
    ]
    feature_name_funcs = (
        {
            "year_of_enrollment_at_cohort_inst": year_of_enrollment_at_cohort_inst,
            "student_has_earned_certificate_at_cohort_inst": ft.partial(
                student_earned_certificate, inst="cohort"
            ),
            "student_has_earned_certificate_at_other_inst": ft.partial(
                student_earned_certificate, inst="other"
            ),
            "term_is_pre_cohort": term_is_pre_cohort,
            "term_is_while_student_enrolled_at_other_inst": term_is_while_student_enrolled_at_other_inst,
            "term_program_of_study_area": term_program_of_study_area,
            "frac_credits_earned": shared.frac_credits_earned,
            "student_term_enrollment_intensity": ft.partial(
                student_term_enrollment_intensity,
                min_num_credits_full_time=min_num_credits_full_time,
            ),
            "num_courses_in_program_of_study_area_term_1": ft.partial(
                num_courses_in_study_area,
                study_area_col="student_program_of_study_area_term_1",
            ),
            "num_courses_in_program_of_study_area_year_1": ft.partial(
                num_courses_in_study_area,
                study_area_col="student_program_of_study_area_year_1",
            ),
            "num_courses_in_term_program_of_study_area": ft.partial(
                num_courses_in_study_area,
                study_area_col="term_program_of_study_area",
            ),
        }
        | {
            fc_col: ft.partial(compute_frac_courses, numer_col=nc_col)
            for nc_col, fc_col in num_frac_courses_cols
        }
        | {
            "frac_sections_students_passed": ft.partial(
                compute_frac_sections_students,
                numer_col="sections_num_students_passed",
            ),
            "frac_sections_students_completed": ft.partial(
                compute_frac_sections_students,
                numer_col="sections_num_students_completed",
            ),
        }
        | {
            "student_pass_rate_above_sections_avg": ft.partial(
                student_rate_above_sections_avg,
                student_col="frac_courses_passed",
                sections_col="frac_sections_students_passed",
            ),
            "student_completion_rate_above_sections_avg": ft.partial(
                student_rate_above_sections_avg,
                student_col="frac_courses_completed",
                sections_col="frac_sections_students_completed",
            ),
        }
    )
    return df.assign(**feature_name_funcs)


def year_of_enrollment_at_cohort_inst(
    df: pd.DataFrame,
    *,
    cohort_start_dt_col: str = "cohort_start_dt",
    term_start_dt_col: str = "term_start_dt",
) -> pd.Series:
    dts_diff = (df[term_start_dt_col].sub(df[cohort_start_dt_col])).dt.days
    return pd.Series(np.ceil((dts_diff + 1) / 365.25), dtype="Int8")


def student_earned_certificate(
    df: pd.DataFrame,
    *,
    inst: t.Literal["cohort", "other"],
    enrollment_year_col: str = "year_of_enrollment_at_cohort_inst",
) -> pd.Series:
    degree_year_cols = [
        f"first_year_to_certificate_at_{inst}_inst",
        f"years_to_latest_certificate_at_{inst}_inst",
    ]
    return df.loc[:, degree_year_cols].lt(df[enrollment_year_col], axis=0).any(axis=1)


def term_is_pre_cohort(
    df: pd.DataFrame,
    *,
    cohort_start_dt_col: str = "cohort_start_dt",
    term_start_dt_col: str = "term_start_dt",
) -> pd.Series:
    return df[term_start_dt_col].lt(df[cohort_start_dt_col]).astype("boolean")


# TODO: we could probably compute this directly, w/o an intermediate feature?
def term_is_while_student_enrolled_at_other_inst(
    df: pd.DataFrame, *, col: str = "num_courses_enrolled_at_other_institution_s_Y"
) -> pd.Series:
    return df[col].gt(0)


def term_program_of_study_area(
    df: pd.DataFrame, *, col: str = "term_program_of_study"
) -> pd.Series:
    return shared.extract_short_cip_code(df[col])


def num_courses_in_study_area(
    df: pd.DataFrame,
    *,
    study_area_col: str,
    course_subject_areas_col: str = "course_subject_areas",
    fill_value: str = "-1",
) -> pd.Series:
    return (
        pd.DataFrame(df[course_subject_areas_col].tolist(), dtype="string")
        .eq(df[study_area_col].fillna(fill_value), axis="index")
        .sum(axis="columns")
        .astype("Int8")
    )


def compute_frac_courses(
    df: pd.DataFrame, *, numer_col: str, denom_col: str = "num_courses"
) -> pd.Series:
    result = df[numer_col].div(df[denom_col])
    if not result.between(0.0, 1.0, inclusive="both").all():
        raise ValueError()
    return result


def compute_frac_sections_students(
    df: pd.DataFrame,
    *,
    numer_col: str,
    denom_col: str = "sections_num_students_enrolled",
) -> pd.Series:
    result = df[numer_col].div(df[denom_col])
    if not result.between(0.0, 1.0, inclusive="both").all():
        raise ValueError()
    return result


def student_rate_above_sections_avg(
    df: pd.DataFrame, *, student_col: str, sections_col: str
) -> pd.Series:
    return df[student_col].gt(df[sections_col])


def student_term_enrollment_intensity(
    df: pd.DataFrame,
    *,
    min_num_credits_full_time: float,
    num_credits_col: str = "num_credits_attempted",
) -> pd.Series:
    if df[num_credits_col].isna().any():
        LOGGER.warning(
            "%s null values found for '%s'; "
            "calculation of student_term_enrollment_intensity doesn't correctly handle nulls",
            df[num_credits_col].isna().sum(),
            num_credits_col,
        )
    return pd.Series(
        data=np.where(
            df[num_credits_col].ge(min_num_credits_full_time), "FULL-TIME", "PART-TIME"
        ),
        index=df.index,
        dtype="string",
    )


def num_courses_col_agg(col: str = "course_id") -> pd.NamedAgg:
    return pd.NamedAgg(col, "count")


def num_courses_passed_col_agg(col: str = "course_passed") -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def num_courses_completed_col_agg(col: str = "course_completed") -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def num_credits_attempted_col_agg(
    col: str = "number_of_credits_attempted",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def num_credits_earned_col_agg(col: str = "number_of_credits_earned") -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def course_ids_col_agg(col: str = "course_id") -> pd.NamedAgg:
    return pd.NamedAgg(col, _agg_values_in_list)


def course_subjects_col_agg(col: str = "course_cip") -> pd.NamedAgg:
    return pd.NamedAgg(col, _agg_values_in_list)


def course_subject_areas_col_agg(col: str = "course_subject_area") -> pd.NamedAgg:
    return pd.NamedAgg(col, _agg_values_in_list)


def _agg_values_in_list(ser: pd.Series) -> list:
    result = ser.tolist()
    assert isinstance(result, list)  # type guard
    return result


def course_id_nunique_col_agg(col: str = "course_id") -> pd.NamedAgg:
    return pd.NamedAgg(col, "nunique")


def course_subject_nunique_col_agg(col: str = "course_cip") -> pd.NamedAgg:
    return pd.NamedAgg(col, "nunique")


def course_subject_area_nunique_col_agg(
    col: str = "course_subject_area",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "nunique")


def course_level_mean_col_agg(col: str = "course_level") -> pd.NamedAgg:
    return pd.NamedAgg(col, "mean")


def course_level_std_col_agg(col: str = "course_level") -> pd.NamedAgg:
    return pd.NamedAgg(col, "std")


def course_grade_numeric_mean_col_agg(col: str = "course_grade_numeric") -> pd.NamedAgg:
    return pd.NamedAgg(col, "mean")


def course_grade_numeric_std_col_agg(col: str = "course_grade_numeric") -> pd.NamedAgg:
    return pd.NamedAgg(col, "std")


def section_num_students_enrolled_mean_col_agg(
    col: str = "section_num_students_enrolled",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "mean")


def section_num_students_enrolled_std_col_agg(
    col: str = "section_num_students_enrolled",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "std")


def sections_num_students_enrolled_col_agg(
    col: str = "section_num_students_enrolled",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def sections_num_students_passed_col_agg(
    col: str = "section_num_students_passed",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def sections_num_students_completed_col_agg(
    col: str = "section_num_students_completed",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def sum_dummy_cols_by_group(
    df: pd.DataFrame, *, grp_cols: list[str], agg_cols: list[str]
) -> pd.DataFrame:
    """
    Compute dummy values for all ``agg_cols`` in ``df`` , then group by ``grp_cols``
    and aggregate by "sum" to get the number of values for each dummy value.

    Args:
        df
        grp_cols
        agg_cols
    """
    return (
        pd.get_dummies(
            df[grp_cols + agg_cols],
            columns=agg_cols,
            sparse=False,
            dummy_na=False,
            drop_first=False,
        )
        .groupby(by=grp_cols, observed=True, as_index=True)
        .agg("sum")
        .rename(columns=_rename_sum_by_group_col)
        .reset_index(drop=False)
    )


def equal_cols_by_group(
    df: pd.DataFrame,
    *,
    grp_cols: list[str],
) -> pd.DataFrame:
    """
    Compute dummy values for all of the num_course features

    Args:
        df
        grp_cols
    """
    num_prefix = constants.NUM_COURSE_FEATURE_COL_PREFIX
    dummy_prefix = constants.DUMMY_COURSE_FEATURE_COL_PREFIX

    course_subject_prefixes = [
        constants.NUM_COURSE_FEATURE_COL_PREFIX + "_course_id",
        constants.NUM_COURSE_FEATURE_COL_PREFIX + "_course_subject_area",
    ]

    dummy_cols = {
        col.replace(num_prefix, dummy_prefix, 1): df[col].ge(1)
        for col in df.columns
        if any(col.startswith(prefix) for prefix in course_subject_prefixes)
    }

    return df.assign(**dummy_cols).reindex(columns=grp_cols + list(dummy_cols.keys()))


def sum_val_equal_cols_by_group(
    df: pd.DataFrame,
    *,
    grp_cols: list[str],
    agg_col_vals: list[tuple[str, t.Any]],
) -> pd.DataFrame:
    """
    Compute equal to specified values for all ``agg_col_vals`` in ``df`` ,
    then group by ``grp_cols`` and aggregate with a "sum".

    Args:
        df
        grp_cols
        agg_col_vals
    """
    temp_col_series = {}
    for col, val in agg_col_vals:
        # make multi-value col names nicer to read
        temp_col = (
            f"{col}_{'|'.join(str(item) for item in val)}"
            if isinstance(val, list)
            else f"{col}_{val}"
        )
        temp_col_series[temp_col] = shared.compute_values_equal(df[col], val)
    return (
        df.assign(**temp_col_series)
        .reindex(columns=grp_cols + list(temp_col_series.keys()))
        .groupby(by=grp_cols, observed=True, as_index=True)
        .agg("sum")
        .rename(columns=_rename_sum_by_group_col)
        .reset_index(drop=False)
    )


def _rename_sum_by_group_col(col: str) -> str:
    return f"{constants.NUM_COURSE_FEATURE_COL_PREFIX}_{col}"


def multicol_grade_aggs_by_group(
    df: pd.DataFrame,
    *,
    min_passing_grade: float,
    grp_cols: list[str],
    grade_col: str = "grade",
    grade_numeric_col: str = "course_grade_numeric",
    section_grade_numeric_col: str = "section_course_grade_numeric_mean",
) -> pd.DataFrame:
    return (
        df.loc[:, grp_cols + [grade_col, grade_numeric_col, section_grade_numeric_col]]
        # compute intermediate column values all at once, which is efficient
        .assign(
            course_grade_is_failing_or_withdrawal=ft.partial(
                _course_grade_is_failing_or_withdrawal,
                min_passing_grade=min_passing_grade,
                grade_col=grade_col,
                grade_numeric_col=grade_numeric_col,
            ),
            course_grade_above_section_avg=ft.partial(
                _course_grade_above_section_avg,
                grade_numeric_col=grade_numeric_col,
                section_grade_numeric_col=section_grade_numeric_col,
            ),
        )
        .groupby(by=grp_cols, observed=True, as_index=False)
        # so that we can efficiently aggregate those intermediate values per group
        .agg(
            num_courses_grade_is_failing_or_withdrawal=(
                "course_grade_is_failing_or_withdrawal",
                "sum",
            ),
            num_courses_grade_above_section_avg=(
                "course_grade_above_section_avg",
                "sum",
            ),
        )
    )


def _course_grade_is_failing_or_withdrawal(
    df: pd.DataFrame,
    min_passing_grade: float,
    grade_col: str = "grade",
    grade_numeric_col: str = "course_grade_numeric",
) -> pd.Series:
    return (
        df[grade_col].isin({"F", "W"})
        | df[grade_numeric_col].between(0.0, min_passing_grade, inclusive="left")
    )  # fmt: skip


def _course_grade_above_section_avg(
    df: pd.DataFrame,
    grade_numeric_col: str = "course_grade_numeric",
    section_grade_numeric_col: str = "section_course_grade_numeric_mean",
) -> pd.Series:
    return df[grade_numeric_col].gt(df[section_grade_numeric_col])



import importlib.metadata
import logging
import os
import shutil
import tempfile
import typing as t
from collections.abc import Collection

import mlflow
import mlflow.artifacts
import mlflow.tracking
import yaml  # type: ignore[import-untyped]

LOGGER = logging.getLogger(__name__)


def load_mlflow_model(
    model_uri: str,
    framework: t.Optional[t.Literal["sklearn", "xgboost", "lightgbm"]] = None,
) -> object:
    """
    Load a (registered) MLFlow model of whichever model type from a specified URI.

    Args:
        model_uri
        framework

    References:
        - https://mlflow.org/docs/latest/python_api/mlflow.sklearn.html#mlflow.sklearn.load_model
        - https://mlflow.org/docs/latest/python_api/mlflow.pyfunc.html#mlflow.pyfunc.load_model
    """
    load_model_func = (
        mlflow.sklearn.load_model
        if framework == "sklearn"
        else mlflow.xgboost.load_model
        if framework == "xgboost"
        else mlflow.lightgbm.load_model
        if framework == "lightgbm"
        else mlflow.pyfunc.load_model
    )
    model = load_model_func(model_uri)
    LOGGER.info("mlflow model loaded from '%s'", model_uri)
    return model


def patch_mlflow_model_env_files(
    run_id: str,
    packages: Collection[str] = ("mlflow", "pandas", "scikit-learn"),
    *,
    force: bool = True,
    mlflow_client: mlflow.tracking.MlflowClient,
) -> None:
    """
    Patch an mlflow model's environment metadata with the versions of packages
    present in the current Python environment.

    Args:
        run_id
        packages
        force
        mlflow_client

    References:
        - https://docs.databricks.com/aws/en/machine-learning/automl/regression-train-api#no-module-named-pandascoreindexesnumeric
    """
    package_versions = {
        package: importlib.metadata.version(package) for package in packages
    }
    # set up local dir for downloading the artifacts
    tmp_dir = str(tempfile.TemporaryDirectory().name)
    os.makedirs(tmp_dir)
    # fix conda.yaml file's pip dependencies
    conda_fpath = mlflow.artifacts.download_artifacts(
        artifact_uri=f"runs:/{run_id}/model/conda.yaml", dst_path=tmp_dir
    )
    with open(conda_fpath) as f:
        conda_env = yaml.load(f, Loader=yaml.FullLoader)
    conda_env["dependencies"][-1]["pip"] = _patch_libs_in_env_file(
        package_versions,
        conda_env["dependencies"][-1]["pip"],
        force=force,
        fname="conda.yaml",
    )
    with open(f"{tmp_dir}/conda.yaml", "w") as f:
        f.write(yaml.dump(conda_env))
    mlflow_client.log_artifact(
        run_id=run_id, local_path=conda_fpath, artifact_path="model"
    )
    # fix requirements.txt file's pip dependencies
    venv_fpath = mlflow.artifacts.download_artifacts(
        artifact_uri=f"runs:/{run_id}/model/requirements.txt", dst_path=tmp_dir
    )
    with open(venv_fpath) as f:
        venv_libs = [lib.strip() for lib in f.readlines()]
    venv_libs = _patch_libs_in_env_file(
        package_versions, venv_libs, force=force, fname="requirements.txt"
    )
    with open(f"{tmp_dir}/requirements.txt", "w") as f:
        f.write("\n".join(venv_libs))
    mlflow_client.log_artifact(
        run_id=run_id, local_path=venv_fpath, artifact_path="model"
    )
    shutil.rmtree(tmp_dir)


def _patch_libs_in_env_file(
    package_versions: dict[str, str], libs: list[str], *, force: bool, fname: str
) -> list[str]:
    for package, version in package_versions.items():
        package_exists = any(lib.startswith(f"{package}==") for lib in libs)
        if not package_exists:
            libs.append(f"{package}=={version}")
            LOGGER.info("adding %s==%s to %s", package, version, fname)
        elif force is True:
            libs = [lib for lib in libs if not lib.startswith(f"{package}==")]
            libs.append(f"{package}=={version}")
            LOGGER.info("overwriting %s==%s in %s", package, version, fname)
    return libs



import numpy as np
import pandas as pd
import pytest

from student_success_tool.preprocessing.features.pdp import cumulative


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        {
            "student_id": ["01", "01", "01", "01", "01"],
            "academic_year": [
                "2020-21",
                "2020-21",
                "2020-21",
                "2021-22",
                "2022-23",
            ],
            "academic_term": ["FALL", "SPRING", "SUMMER", "FALL", "FALL"],
            "term_id": [
                "2020-21 FALL",
                "2020-21 SPRING",
                "2020-21 SUMMER",
                "2021-22 FALL",
                "2022-23 FALL",
            ],
            "course_grade_num_mean": [4.0, 2.5, 1.75, 3.25, 3.5],
            "num_courses": [3, 2, 2, 2, 1],
            "num_courses_course_level_0": [2, 1, 0, 0, 0],
            "num_courses_course_level_1": [1, 1, 2, 2, 1],
            "took_course_subject_area_02": [True, False, False, False, True],
            "took_course_subject_area_05_06": [False, False, False, True, False],
            "num_credits_earned": [3, 12, 15, 4, 15],
            "course_ids": [
                ["A101", "B101", "C201"],
                ["A101", "D201"],
                ["E201", "F201"],
                ["G201", "H201"],
                ["H201"],
            ],
            "course_subject_areas": [
                ["01", "02", "03"],
                ["01", "04"],
                ["03", "04"],
                ["05", "06"],
                ["02"],
            ],
            "term_rank": [1, 3, 4, 5, 9],
            "term_rank_core": [1, 2, pd.NA, 3, 5],
            "term_rank_noncore": [pd.NA, pd.NA, 2, pd.NA, pd.NA],
            "term_is_pre_cohort": [True, False, False, False, False],
        }
    ).astype(
        {"term_rank": "Int8", "term_rank_core": "Int8", "term_rank_noncore": "Int8"}
    )


@pytest.fixture(scope="module")
def df_grped(df):
    return (
        df.sort_values(by=["student_id", "academic_year", "academic_term"])
        .groupby(by="student_id", as_index=True, observed=True, sort=False)
    )  # fmt: skip


@pytest.mark.parametrize(
    ["num_course_cols", "dummy_course_cols", "col_aggs", "exp"],
    [
        (
            [
                "num_courses_course_level_0",
                "num_courses_course_level_1",
                "num_credits_earned",
            ],
            ["took_course_subject_area_02", "took_course_subject_area_05_06"],
            [
                ("term_id", "count"),
                ("term_is_pre_cohort", "sum"),
                ("course_grade_num_mean", ["mean", "std"]),
                ("num_courses", "sum"),
            ],
            pd.DataFrame(
                {
                    "cumcount_term_id": [1.0, 2.0, 3.0, 4.0, 5.0],
                    "cumsum_term_is_pre_cohort": [1.0, 1.0, 1.0, 1.0, 1.0],
                    "cummean_course_grade_num_mean": [4.0, 3.25, 2.75, 2.875, 3.0],
                    "cumstd_course_grade_num_mean": [
                        np.nan,
                        1.061,
                        1.146,
                        0.968,
                        0.884,
                    ],
                    "cumsum_num_courses": [3.0, 5.0, 7.0, 9.0, 10.0],
                    "cumsum_num_credits_earned": [3.0, 15.0, 30.0, 34.0, 49.0],
                    "cummax_took_course_subject_area_02": [1.0, 1.0, 1.0, 1.0, 1.0],
                    "cummax_took_course_subject_area_05_06": [
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        1.0,
                    ],
                    "cumfrac_num_courses_course_level_0": [
                        0.667,
                        0.6,
                        0.429,
                        0.333,
                        0.3,
                    ],
                    "cumfrac_num_courses_course_level_1": [
                        0.333,
                        0.4,
                        0.571,
                        0.667,
                        0.7,
                    ],
                    "cummax_in_12_creds_took_course_subject_area_02": [
                        True,
                        True,
                        True,
                        True,
                        True,
                    ],
                    "cummax_in_12_creds_took_course_subject_area_05_06": [
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                }
            ),
        ),
    ],
)
def test_expanding_agg_features(
    df_grped, num_course_cols, dummy_course_cols, col_aggs, exp
):
    obs = cumulative.expanding_agg_features(
        df_grped,
        num_course_cols=num_course_cols,
        col_aggs=col_aggs,
        dummy_course_cols=dummy_course_cols,
        credits=12,
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    # raises error if not equal
    print(obs.columns)
    print(exp.columns)
    assert pd.testing.assert_frame_equal(obs, exp, rtol=0.001) is None


@pytest.mark.parametrize(
    ["cols", "exp"],
    [
        (
            ["course_ids", "course_subject_areas"],
            pd.DataFrame(
                {
                    "cumnum_unique_course_ids": [3, 4, 6, 8, 8],
                    "cumnum_unique_course_subject_areas": [3, 4, 4, 6, 6],
                    "cumnum_repeated_course_ids": [0, 1, 1, 1, 2],
                    "cumnum_repeated_course_subject_areas": [0, 1, 3, 3, 4],
                },
                dtype="Int16",
            ),
        ),
    ],
)
def test_cumnum_unique_and_repeated_features(df_grped, cols, exp):
    obs = cumulative.cumnum_unique_and_repeated_features(df_grped, cols=cols)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    # raises error if not equal
    assert pd.testing.assert_frame_equal(obs, exp, rtol=0.001) is None


@pytest.mark.parametrize(
    "exp_new",
    [
        pd.DataFrame(
            {
                "min_student_term_rank": [1, 1, 1, 1, 1],
                "min_student_term_rank_core": [1, 1, 1, 1, 1],
                "min_student_term_rank_noncore": [2, 2, 2, 2, 2],
                "cumfrac_terms_enrolled": [1.0, 0.666, 0.75, 0.8, 0.556],
                "cumfrac_core_terms_enrolled": [1.0, 1.0, pd.NA, 1.0, 0.8],
                "cumfrac_noncore_terms_enrolled": [pd.NA, pd.NA, 1.0, pd.NA, pd.NA],
            }
        ).astype(
            {
                "min_student_term_rank": "Int8",
                "min_student_term_rank_core": "Int8",
                "min_student_term_rank_noncore": "Int8",
                "cumfrac_terms_enrolled": "Float32",
                "cumfrac_core_terms_enrolled": "Float32",
                "cumfrac_noncore_terms_enrolled": "Float32",
            }
        ),
    ],
)
def test_add_cumfrac_terms_enrolled_features(df, exp_new):
    # HACK: let's add the couple cumulative feature pre-requisites here
    df = df.assign(
        cumnum_terms_enrolled=pd.Series([1.0, 2.0, 3.0, 4.0, 5.0]),
        cumnum_core_terms_enrolled=pd.Series([1.0, 2.0, 2.0, 3.0, 4.0]),
        cumnum_noncore_terms_enrolled=pd.Series([0.0, 0.0, 1.0, 1.0, 1.0]),
    )
    obs = cumulative.add_cumfrac_terms_enrolled_features(
        df, student_id_cols=["student_id"]
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    # all *existing* columns are unchanged
    assert obs.loc[:, df.columns].equals(df)
    # all *new* columns are as expected
    assert (
        pd.testing.assert_frame_equal(obs.drop(columns=df.columns), exp_new, rtol=0.01)
        is None
    )


@pytest.mark.parametrize(
    ["cols", "max_term_num", "exp_new"],
    [
        (
            ["num_courses", "course_grade_num_mean"],
            3,
            pd.DataFrame(
                {
                    "num_courses_diff_prev_term": [np.nan, -1.0, 0.0, 0.0, -1.0],
                    "course_grade_num_mean_diff_prev_term": [
                        np.nan,
                        -1.5,
                        -0.75,
                        1.5,
                        0.25,
                    ],
                    "num_courses_diff_term_1_to_term_2": [-1.0, -1.0, -1.0, -1.0, -1.0],
                    "num_courses_diff_term_2_to_term_3": [0.0, 0.0, 0.0, 0.0, 0.0],
                    "course_grade_num_mean_diff_term_1_to_term_2": [
                        -1.5,
                        -1.5,
                        -1.5,
                        -1.5,
                        -1.5,
                    ],
                    "course_grade_num_mean_diff_term_2_to_term_3": [
                        -0.75,
                        -0.75,
                        -0.75,
                        -0.75,
                        -0.75,
                    ],
                }
            ),
        ),
        (
            ["num_courses"],
            2,
            pd.DataFrame(
                {
                    "num_courses_diff_prev_term": [np.nan, -1.0, 0.0, 0.0, -1.0],
                    "num_courses_diff_term_1_to_term_2": [-1.0, -1.0, -1.0, -1.0, -1.0],
                }
            ),
        ),
    ],
)
def test_add_term_diff_features(df, cols, max_term_num, exp_new):
    # HACK: let's add the couple cumulative feature pre-requisites here
    df = df.assign(
        cumnum_terms_enrolled=pd.Series([1.0, 2.0, 3.0, 4.0, 5.0]),
        cumnum_core_terms_enrolled=pd.Series([1.0, 2.0, 2.0, 3.0, 4.0]),
    )
    obs = cumulative.add_term_diff_features(
        df,
        cols=cols,
        max_term_num=max_term_num,
        student_id_cols=["student_id"],
        term_num_col="cumnum_terms_enrolled",
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    # all *existing* columns are unchanged
    assert obs.loc[:, df.columns].equals(df)
    # all *new* columns are as expected
    assert (
        pd.testing.assert_frame_equal(obs.drop(columns=df.columns), exp_new, rtol=0.01)
        is None
    )



import logging
import typing as t

import pandas as pd

LOGGER = logging.getLogger(__name__)


def rename_mangled_column_names(
    df: pd.DataFrame, *, overrides: t.Optional[dict[str, str]] = None
) -> pd.DataFrame:
    """
    Rename column names in ``df`` that have been mangled somewhere in the data generation
    and/or ingestion process, so they match the standard data schema.

    Args:
        df: Raw cohort dataset
        overrides: Mapping of raw (mangled) to raw (valid) column names in ``df``
            to be renamed. Note that repeatedly mangled columns are already included,
            so only specify this if the base set doesn't cover a school's situation.

    See Also:
        :class:`schemas.pdp.RawPDPCohortDataSchema`
    """
    rename_columns = {
        "attemptedgatewaymathyear_1": "attempted_gateway_math_year_1",
        "attemptedgatewayenglishyear_1": "attempted_gateway_english_year_1",
        "completedgatewaymathyear_1": "completed_gateway_math_year_1",
        "completedgatewayenglishyear_1": "completed_gateway_english_year_1",
        "gatewaymathgradey_1": "gateway_math_grade_y_1",
        "gatewayenglishgradey_1": "gateway_english_grade_y_1",
        "attempteddevmathy_1": "attempted_dev_math_y_1",
        "attempteddevenglishy_1": "attempted_dev_english_y_1",
        "completeddevmathy_1": "completed_dev_math_y_1",
        "completeddevenglishy_1": "completed_dev_english_y_1",
    }
    if overrides:
        rename_columns |= overrides
    # only rename mangled columns if they actually exist in the data
    rename_columns = {
        old_col: new_col
        for old_col, new_col in rename_columns.items()
        if old_col in df.columns
    }
    LOGGER.info("renaming mangled columns in raw cohort data: %s", rename_columns)
    return df.rename(columns=rename_columns)
